#!/usr/bin/env python

import argparse

import dendropy

# See https://jeetsukumaran.github.io/DendroPy/schemas/index.html for the available
# formats.
FORMATS = {"newick", "nexml", "nexus", "phylip"}


def parseArgs():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Print information phylogenetic tree nodes.",
    )

    parser.add_argument(
        "treeFile",
        help="The tree file to examine.",
    )

    parser.add_argument(
        "--format",
        default="newick",
        choices=sorted(FORMATS),
        help="The input tree format.",
    )

    parser.add_argument(
        "--indent",
        type=int,
        default=2,
        help=(
            "The number of spaces to increase node indentation by when descending to "
            "children."
        ),
    )

    parser.add_argument(
        "--lengths",
        action="store_true",
        help="Also show edge lengths.",
    )

    parser.add_argument(
        "--rootNode",
        help="Re-root the tree on the edge leading into the tip node with this name.",
    )

    parser.add_argument(
        "--noTaxon",
        default="NO-TAXON!",
        help="What to print if a tip node has no taxon name.",
    )

    parser.add_argument(
        "--noLabel",
        default="NO-LABEL!",
        help="What to print if an internal node has no label.",
    )

    return parser.parse_args()


def main() -> None:
    """
    Print the tree by calling the recursive node printer on the root node.
    """
    args = parseArgs()

    kwargs = (
        dict(preserve_underscores=True) if args.format in {"newick", "nexus"} else {}
    )

    with open(args.treeFile) as fp:
        tree = dendropy.Tree.get(file=fp, schema=args.format, **kwargs)

    if args.rootNode:
        node = tree.find_node_with_taxon_label(args.rootNode)
        if not node:
            exit(f"Could not find tree node {args.rootNode!r} in {args.treeFile!r}.")
        tree.reroot_at_edge(node.edge, update_bipartitions=False)

    counts = dict.fromkeys(("tips", "internals"), 0)
    printNode(
        tree.seed_node, 0, args.indent, args.lengths, args.noTaxon, args.noLabel, counts
    )

    totalNodes = counts["tips"] + counts["internals"]
    print(
        f"The tree has {totalNodes} nodes ({counts['internals']} "
        f"internal, {counts['tips']} tips).",
    )


def printNode(
    node: dendropy.Node,
    depth: int,
    indent: int,
    showLengths: bool,
    noTaxon: str,
    noLabel: str,
    counts: dict[str, int],
) -> None:
    """
    Print a single node and (recursively) its descendants.

    @param node: The node to print.
    @param depth: The distance this node is from the root.
    @param indent: The number of spaces to indent the output by for each step away from
        the root.
    @param showLengths: If true, also print edge lengths.
    @param counts: A C{dict} with 'internals' and 'tips' keys used to accumulate the
        count of each node type during the recursive calls.
    """
    line = [f"{' ' * depth * indent} {depth}:"]

    length = node.edge.length if showLengths else ""
    if node.is_internal():
        counts["internals"] += 1
        children = node.child_nodes()
        line.append(f"Int {node.label or noLabel} {length}")
        if depth == 0:
            c = "child" if len(children) == 1 else "children"
            line.append(f"(root has {len(children)} {c})")

        print(" ".join(line))

        for child in children:
            printNode(child, depth + 1, indent, showLengths, noTaxon, noLabel, counts)

    else:
        counts["tips"] += 1
        line.append(
            f"Tip {noTaxon if node.taxon is None else node.taxon.label} {length}"
        )
        print(" ".join(line))


if __name__ == "__main__":
    main()



from unittest import TestCase

from dark.colors import ColorsForCounts


class TestColorsForCounts(TestCase):
    """
    Test the ColorsForCounts class.
    """

    def testCountNotInt(self):
        """
        A C{ValueError} must be raised if a non-integer count is given.
        """
        error = (
            r"^color arguments must be given as space-separated pairs of "
            r"\"count color\" where the count is an integer threshold\. "
            r"Your value \('non-int'\) was not an integer\.$"
        )
        self.assertRaisesRegex(ValueError, error, ColorsForCounts, ("non-int red",))

    def testCountNegative(self):
        """
        A C{ValueError} must be raised if a count is less than zero.
        """
        error = (
            r"^color arguments must be given as space-separated pairs "
            r"of \"count color\" where the count is non-negative. Your "
            r"value \(-4\) is less than 0\.$"
        )
        self.assertRaisesRegex(ValueError, error, ColorsForCounts, ("-4 red",))

    def testNoSpace(self):
        """
        A C{ValueError} must be raised if a count/color value has no space.
        """
        error = (
            r"^color arguments must be given as space-separated pairs "
            r"of \"value color\"\. Your value \('4red'\) does not "
            r"contain a space\.$"
        )
        self.assertRaisesRegex(ValueError, error, ColorsForCounts, ("4red",))

    def testRepeatedCount(self):
        """
        A C{ValueError} must be raised if a count is repeated.
        """
        error = r"^repeated color argument count \(4\)\.$"
        self.assertRaisesRegex(ValueError, error, ColorsForCounts, ("4 red", "4 black"))

    def testThresholdForCountNegative(self):
        """
        The thresholdForCount method must raise AssertionError if the count
        is negative.
        """
        colors = ColorsForCounts(("4 red",))
        error = r"^Count \(-4\) cannot be negative\.$"
        self.assertRaisesRegex(AssertionError, error, colors.thresholdForCount, -4)

    def testThresholdForTooLowCount(self):
        """
        The thresholdForCount method must work correctly when the lowest
        threshold is not reached.
        """
        colors = ColorsForCounts(("4 red",))
        self.assertEqual(0, colors.thresholdForCount(2))

    def testThresholdForCountOneThreshold(self):
        """
        The thresholdForCount method must work correctly when there is one
        threshold.
        """
        colors = ColorsForCounts(("4 red",))
        self.assertEqual(4, colors.thresholdForCount(4))
        self.assertEqual(4, colors.thresholdForCount(10))

    def testThresholdForCountTwoThresholds(self):
        """
        The thresholdForCount method must work correctly when there are two
        thresholds.
        """
        colors = ColorsForCounts(("4 red", "10 green"))
        self.assertEqual(4, colors.thresholdForCount(6))
        self.assertEqual(10, colors.thresholdForCount(12))

    def testThresholdForCountThreeThresholds(self):
        """
        The thresholdForCount method must work correctly when there are three
        thresholds.
        """
        colors = ColorsForCounts(("4 red", "10 green", "100 blue"))
        self.assertEqual(4, colors.thresholdForCount(6))
        self.assertEqual(10, colors.thresholdForCount(86))
        self.assertEqual(100, colors.thresholdForCount(1000))

    def testColorForTooLowCount(self):
        """
        The colorForCount method must work correctly when the lowest
        threshold is not reached.
        """
        colors = ColorsForCounts(("4 red",))
        self.assertEqual("black", colors.colorForCount(2))

    def testColorForTooLowCountDefault(self):
        """
        The colorForCount method must work correctly when the lowest
        threshold is not reached and a default color is given.
        """
        colors = ColorsForCounts(("4 red",), defaultColor="fuschia")
        self.assertEqual("fuschia", colors.colorForCount(2))

    def testColorForCountOneThreshold(self):
        """
        The colorForCount method must work correctly when there is one
        threshold.
        """
        colors = ColorsForCounts(("4 red",))
        self.assertEqual("black", colors.colorForCount(1))
        self.assertEqual("red", colors.colorForCount(4))
        self.assertEqual("red", colors.colorForCount(10))

    def testColorForCountTwoThresholds(self):
        """
        The colorForCount method must work correctly when there are two
        thresholds.
        """
        colors = ColorsForCounts(("4 red", "10 green"))
        self.assertEqual("black", colors.colorForCount(1))
        self.assertEqual("red", colors.colorForCount(6))
        self.assertEqual("green", colors.colorForCount(12))

    def testColorForCountThreeThresholds(self):
        """
        The colorForCount method must work correctly when there are three
        thresholds.
        """
        colors = ColorsForCounts(("4 red", "10 green", "100 blue"))
        self.assertEqual("black", colors.colorForCount(1))
        self.assertEqual("red", colors.colorForCount(6))
        self.assertEqual("green", colors.colorForCount(86))
        self.assertEqual("blue", colors.colorForCount(1000))

    def testThresholdToCssName(self):
        """
        The thresholdToCssName method must work as expected.
        """
        colors = ColorsForCounts(("4 red",))
        self.assertEqual("threshold-4", colors.thresholdToCssName(4))

    def testColorsForThreeThresholds(self):
        """
        The colors attribute must be as expected when there are three
        thresholds.
        """
        colors = ColorsForCounts(("4 red", "10 green", "100 blue"), defaultColor="pink")
        self.assertEqual(
            ((100, "blue"), (10, "green"), (4, "red"), (0, "pink")), colors.colors
        )



def printBlastRecordForDerek(record):
    """
    This is a hacked version of printBlastRecord that I wrote in
    Addenbrookes hospital so Derek could try hacking on the reads
    that had low e values.  It's called by bin/print-blast-xml-for-derek.py
    """
    if len(record.alignments) == 0:
        return 0
    if record.descriptions[0].e > 1e-120:
        return 0
    print(record.query, end=" ")
    for i, alignment in enumerate(record.alignments):
        print(record.descriptions[i].e, end=" ")
        for hspIndex, hsp in enumerate(alignment.hsps, start=1):
            if hsp.sbjct_start < hsp.sbjct_end:
                sense = 1
                if hsp.sbjct_start < 2200:
                    side = "left"
                else:
                    side = "right"
            else:
                sense = -1
                if hsp.sbjct_end < 2200:
                    side = "left"
                else:
                    side = "right"
            print(sense, side)
            break
    return 1



import sys
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path
from typing import BinaryIO, TextIO

if sys.version_info < (3, 10):
    raise Exception("The dark matter code needs Python 3.10 or later.")  # pyright: ignore[reportUnreachable]


try:
    __version__ = version("dark-matter")
except PackageNotFoundError:
    # Package is not installed.
    __version__ = "unknown"


File = TextIO | BinaryIO | str | bytes | Path



#!/usr/bin/env python

import argparse

from dark.proteins import SqliteIndex

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=(
        "Query an sqlite3 protein/genome database and write "
        "the matching sequences to standard output as FASTA."
    ),
)

parser.add_argument(
    "--databaseFile",
    required=True,
    help=(
        "The protein/genome database file. This must be an sqlite3 database "
        "file created by make-protein-database.py"
    ),
)

parser.add_argument(
    "--query",
    help=(
        "An SQL database query. "
        "The query must result in rows with two columns, the sequence id "
        "and the sequence. These can be protein or nucleotide sequences, "
        "but it is probably better not to mix them!"
    ),
)

parser.add_argument(
    "--queryFile",
    help=(
        "The name of a file containing an SQL database query. "
        "The query must result in rows with two columns, the sequence id "
        "and the sequence. These can be protein or nucleotide sequences, "
        "but it is probably better not to mix them!"
    ),
)

parser.add_argument(
    "--keepDuplicateIds",
    default=False,
    action="store_true",
    help="If specified, do not remove duplicate ids.",
)

parser.add_argument(
    "--idsOnly",
    default=False,
    action="store_true",
    help="If specified, only print matching sequence ids.",
)

args = parser.parse_args()

db = SqliteIndex(args.databaseFile)

if args.queryFile:
    query = open(args.queryFile).read()
elif args.query:
    query = args.query
else:
    raise RuntimeError("You must use either --query or --queryFile")

cur = db.execute(query)

removeDuplicates = not args.keepDuplicateIds
idsOnly = args.idsOnly

if removeDuplicates:
    seen = set()

while True:
    rows = cur.fetchmany()
    if rows:
        for row in rows:
            try:
                id_, sequence, genomeName = row
            except ValueError:
                id_, sequence = row
            else:
                id_ = "%s [%s]" % (id_, genomeName)

            if removeDuplicates:
                if id_ in seen:
                    continue
                else:
                    seen.add(id_)
            if idsOnly:
                print(id_)
            else:
                print(">%s\n%s" % (id_, sequence))
    else:
        break



#!/usr/bin/env python

import argparse
import sys
from itertools import chain

from dark.fasta import FastaReads
from dark.fastq import FastqReads

if __name__ == "__main__":
    # We do not use the addFASTACommandLineOptions and
    # parseFASTACommandLineOptions utility functions below because we allow
    # multiple FASTA or FASTQ files on the command line, which we specify
    # by --fasta and --fastq. And those names clash with the option names
    # used by those utility functions.

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Filter reads alignments to select a subset of reads. ",
            "Given BLAST or DIAMOND JSON output files and the "
            "FASTA (or FASTQ) query sequence files given to BLAST or "
            "DIAMOND, and some filtering criteria, print the input "
            "reads that pass the filtering.",
        ),
    )

    parser.add_argument(
        "--matcher",
        default="blast",
        choices=("blast", "diamond"),
        help="The matching algorithm that was used to produce the JSON.",
    )

    parser.add_argument(
        "--json",
        metavar="FILE.JSON[.BZ2]",
        nargs="+",
        action="append",
        required=True,
        help="the JSON file(s) of BLAST or DIAMOND output.",
    )

    # A mutually exclusive group for either FASTA or FASTQ files.
    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "--fasta",
        metavar="FILE.FASTA",
        nargs="+",
        action="append",
        help=("the FASTA file(s) of sequences that were given to BLAST or DIAMOND."),
    )

    group.add_argument(
        "--fastq",
        metavar="FILE.FASTQ",
        nargs="+",
        action="append",
        help=("the FASTQ file(s) of sequences that were given to BLAST or DIAMOND."),
    )

    # Args specific to DIAMOND.

    # A group for either the DIAMOND FASTA file or a sqlite3 database
    # of the FASTA used to make the DIAMOND database.
    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        "--diamondDatabaseFastaFilename",
        metavar="FILE.FASTA",
        help=(
            "The filename of the FASTA file used to make the DIAMOND "
            "database. If --matcher diamond is used, either this argument "
            "or --diamondSqliteDatabaseFilename must be specified."
        ),
    )

    group.add_argument(
        "--diamondSqliteDatabaseFilename",
        metavar="FILE.SQL",
        help=(
            "The filename of the sqlite3 database file of FASTA metadata, "
            "made from the FASTA that was used to make the DIAMOND "
            "database. If --matcher diamond is used, either this argument "
            "or --diamondDatabaseFilename must be specified."
        ),
    )

    parser.add_argument(
        "--diamondDatabaseFastaDirectory",
        metavar="DIR",
        help=(
            "The directory where the FASTA file used to make the DIAMOND "
            "database can be found. This argument is only useful when "
            "--diamondSqliteDatabaseFilename is specified."
        ),
    )

    # Args for filtering on ReadsAlignments.
    parser.add_argument(
        "--minStart",
        type=int,
        metavar="OFFSET",
        help="Reads that start before this subject offset should not be shown.",
    )

    parser.add_argument(
        "--maxStop",
        type=int,
        metavar="OFFSET",
        help="Reads that end after this subject offset should not be shown.",
    )

    parser.add_argument(
        "--oneAlignmentPerRead",
        default=False,
        action="store_true",
        help="If True, only keep the best alignment for each read.",
    )

    parser.add_argument(
        "--maxAlignmentsPerRead",
        type=int,
        metavar="N",
        help=(
            "Reads with more than this many alignments will be elided. Pass "
            "zero to only keep reads with no matches (alignments)."
        ),
    )

    parser.add_argument(
        "--scoreCutoff",
        type=float,
        metavar="SCORE",
        help=("A float score. Matches with scores worse than this will be ignored."),
    )

    parser.add_argument(
        "--maxHspsPerHit",
        type=int,
        metavar="N",
        help=("A numeric maximum number of HSPs to keep for each subject match."),
    )

    parser.add_argument(
        "--whitelist",
        nargs="+",
        action="append",
        metavar="TITLE",
        help="subject titles that should be whitelisted",
    )

    parser.add_argument(
        "--blacklist",
        nargs="+",
        action="append",
        metavar="TITLE",
        help="subject titles that should be blacklisted",
    )

    parser.add_argument(
        "--titleRegex",
        metavar="TITLE-REGEX",
        help="a regex that subject titles must match.",
    )

    parser.add_argument(
        "--negativeTitleRegex",
        metavar="TITLE-REGEX",
        help="a regex that subject titles must not match.",
    )

    parser.add_argument(
        "--truncateTitlesAfter",
        help=(
            "a string that subject titles will be truncated beyond. If the "
            "truncated version of a title has already been seen, "
            "that title will be skipped."
        ),
    )

    parser.add_argument(
        "--minSequenceLen",
        type=int,
        metavar="N",
        help="subjects of lesser length will be elided.",
    )

    parser.add_argument(
        "--maxSequenceLen",
        type=int,
        metavar="N",
        help="subjects of greater length will be elided.",
    )

    parser.add_argument(
        "--taxonomy",
        metavar="NAME",
        help=(
            "the taxonomic group that subjects must match "
            'E.g., "Vira" will filter on viruses.'
        ),
    )

    args = parser.parse_args()

    # Flatten lists of lists that we get from using both nargs='+' and
    # action='append'. We use both because it allows people to use (e.g.)
    # --json on the command line either via "--json file1 --json file2" or
    # "--json file1 file2", or a combination of these. That way it's not
    # necessary to remember which way you're supposed to use it and you also
    # can't be hit by the subtle problem encountered in
    # https://github.com/acorg/dark-matter/issues/453
    jsonFiles = list(chain.from_iterable(args.json))
    whitelist = set(chain.from_iterable(args.whitelist)) if args.whitelist else None
    blacklist = set(chain.from_iterable(args.blacklist)) if args.blacklist else None

    # TODO: Add a --readClass command-line option in case we want to
    # process FASTA containing AA sequences.
    if args.fasta:
        reads = FastaReads(list(chain.from_iterable(args.fasta)))
    else:
        reads = FastqReads(list(chain.from_iterable(args.fastq)))

    if args.matcher == "blast":
        from dark.blast.alignments import BlastReadsAlignments

        readsAlignments = BlastReadsAlignments(reads, jsonFiles)
    else:
        # Must be 'diamond' (due to parser.add_argument 'choices' argument).
        if (
            args.diamondDatabaseFastaFilename is None
            and args.diamondSqliteDatabaseFilename is None
        ):
            print(
                "Either --diamondDatabaseFastaFilename or "
                "--diamondSqliteDatabaseFilename must be used with "
                "--matcher diamond.",
                file=sys.stderr,
            )
            sys.exit(1)
        elif not (
            args.diamondDatabaseFastaFilename is None
            or args.diamondSqliteDatabaseFilename is None
        ):
            print(
                "--diamondDatabaseFastaFilename and "
                "--diamondSqliteDatabaseFilename cannot both be used with "
                "--matcher diamond.",
                file=sys.stderr,
            )
            sys.exit(1)

        from dark.diamond.alignments import DiamondReadsAlignments

        readsAlignments = DiamondReadsAlignments(
            reads,
            jsonFiles,
            databaseFilename=args.diamondDatabaseFastaFilename,
            databaseDirectory=args.diamondDatabaseFastaDirectory,
            sqliteDatabaseFilename=args.diamondSqliteDatabaseFilename,
        )

    readsAlignments.filter(
        maxAlignmentsPerRead=args.maxAlignmentsPerRead,
        minSequenceLen=args.minSequenceLen,
        maxSequenceLen=args.maxSequenceLen,
        minStart=args.minStart,
        maxStop=args.maxStop,
        oneAlignmentPerRead=args.oneAlignmentPerRead,
        maxHspsPerHit=args.maxHspsPerHit,
        scoreCutoff=args.scoreCutoff,
        whitelist=whitelist,
        blacklist=blacklist,
        titleRegex=args.titleRegex,
        negativeTitleRegex=args.negativeTitleRegex,
        truncateTitlesAfter=args.truncateTitlesAfter,
        taxonomy=args.taxonomy,
    )

    format_ = "fasta" if args.fasta else "fastq"
    write = sys.stdout.write

    for readAlignments in readsAlignments:
        write(readAlignments.read.toString(format_=format_))



#!/usr/bin/env python

import argparse

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions

parser = argparse.ArgumentParser(
    description=(
        "Given FASTA/Q on stdin, write FASTQ to stdout with all nucleotide quality "
        "scores set to a given value."
    )
)

parser.add_argument(
    "--quality",
    "-q",
    type=int,
    default=30,
    help="The integer PHRED quality to set. Sanger (32-based PHRED is used).",
)

addFASTACommandLineOptions(parser)
args = parser.parse_args()
reads = parseFASTACommandLineOptions(args)
quality = chr(ord(" ") + args.quality)

for read in reads:
    read.quality = quality * len(read)
    print(read.toString(format_="fastq"), end="")



#!/usr/bin/env python

import sys

from pysam import AlignmentFile

from dark.filter import (
    addFASTAFilteringCommandLineOptions,
    parseFASTAFilteringCommandLineOptions,
)
from dark.reads import Reads
from dark.sam import SAMFilter, samfile

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Given a SAM/BAM file and a set of filtering criteria "
            "write filtered SAM/BAM to stdout."
        ),
    )

    parser.add_argument(
        "--quiet",
        action="store_true",
        default=False,
        help="If True, do not print the final summary.",
    )

    parser.add_argument(
        "--bam",
        action="store_const",
        const="b",
        default="",
        help="If given, write (gzip compressed) BAM output.",
    )

    parser.add_argument(
        "--checkResultCount",
        type=int,
        help=(
            "The number of alignments expected in the output. If this "
            "number is not seen, the script exits with status 1 (and an "
            "error message is also printed, unless --quiet was used)."
        ),
    )

    addFASTAFilteringCommandLineOptions(parser)
    SAMFilter.addFilteringOptions(parser)

    args = parser.parse_args()
    reads = parseFASTAFilteringCommandLineOptions(args, Reads())
    samFilter = SAMFilter.parseFilteringOptions(
        args, reads.filterRead, storeQueryIds=True
    )

    # The following 'if' has a False in it to make it always fail. That's
    # because pysam issue 716 (see below) did not fix the problem as I had
    # hoped. Instead it throws an error if you pass a header that has a
    # modified SQ key with reference ids and there's a difference it
    # doesn't like. It's always safe to use the 'else' below, with the
    # slight downside being that its header will mention all sequence ids,
    # even if you only want a lesser number (via --referenceId). I'm
    # leaving the code here because this is how you would do it, and it
    # might be possible to just copy the 'header' dict below and further
    # adjust it to avoid the pysam error.
    if False and samFilter.referenceIds:
        # Make a header that only includes the wanted reference ids (if
        # any).
        print(
            "If you get a segmentation violation, this is (I think) "
            "due to the following pysam issue "
            "https://github.com/pysam-developers/pysam/issues/716 It is "
            "safer to run without using --referenceId.",
            file=sys.stderr,
        )
        with samfile(args.samfile) as sam:
            header = sam.header.as_dict()
        referenceIds = samFilter.referenceIds
        sequences = [
            sequence for sequence in header["SQ"] if sequence["SN"] in referenceIds
        ]
        referenceIdsFound = set(sequence["SN"] for sequence in sequences)
        notFound = referenceIds - referenceIdsFound
        if notFound:
            if len(notFound) == len(referenceIds):
                raise ValueError(
                    "None of the provided --referenceId ids were found in %s"
                    % args.samfile
                )
            else:
                # Just warn about the missing references.
                print(
                    "WARNING: reference%s (%s) given with --referenceId not "
                    "found in %s."
                    % (
                        "" if len(notFound) == 1 else "s",
                        ", ".join(sorted(notFound)),
                        args.samfile,
                    ),
                    file=sys.stderr,
                )

        header["SQ"] = sequences
        out = AlignmentFile(sys.stdout, mode="w" + args.bam, header=header)
    else:
        with samfile(args.samfile) as sam:
            out = AlignmentFile(sys.stdout, mode="w" + args.bam, template=sam)

    save = out.write
    kept = 0
    for kept, alignment in enumerate(samFilter.alignments(), start=1):
        save(alignment)

    out.close()

    if not args.quiet:
        total = samFilter.alignmentCount
        print(
            "Read %d alignment%s, kept %d (%.2f%%)."
            % (
                total,
                "" if total == 1 else "s",
                kept,
                0.0 if total == 0 else kept / total * 100.0,
            ),
            file=sys.stderr,
        )

    if args.checkResultCount is not None and kept != args.checkResultCount:
        if not args.quiet:
            print(
                "Did not write the expected %d alignment%s (wrote %d)."
                % (
                    args.checkResultCount,
                    "" if args.checkResultCount == 1 else "s",
                    kept,
                ),
                file=sys.stderr,
            )
        sys.exit(1)



#!/usr/bin/env python

"""
Read DNA FASTA from stdin and print FASTA to stdout, only including sequences
that match the requested taxonomy regular expression at any level.

This script will only produce output if the FASTA it is given has description
lines like

  >gi|9630656|ref|NC_001925.1| Bunyamwera virus L segment, complete sequence

with a GI number in the second '|'-separated field. This is the format found
in the NCBI reference sequence files. E.g. at
ftp://ftp.ncbi.nih.gov/refseq/release/viral

You must have the NCBI taxonomy database installed locally for this code to
succeed.  See the file doc/taxonomy.md for instructions on how to set that up.
If you see sequences unexpectedly rejected because they have no associated
taxonomy, make sure you have the latest taxonomy files loaded into MySQL.
"""

import argparse
import re
import sys
from typing import Optional, TextIO

from dark.fasta import FastaReads
from dark.taxonomy import LineageFetcher


def writeDetails(accept, readId, taxonomy, fp):
    """
    Write read and taxonomy details.

    @param accept: A C{bool} indicating whether the read was accepted,
        according to its taxonomy.
    @param readId: The C{str} id of the read.
    @taxonomy: A C{list} of taxonomy C{str} levels.
    @fp: An open file pointer to write to.
    """
    fp.write(
        "%s %s\n       %s\n\n"
        % (
            "MATCH:" if accept else "MISS: ",
            readId,
            " | ".join(taxonomy) if taxonomy else "No taxonomy found.",
        )
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Filter FASTA based on taxonomy",
        epilog="Read DNA FASTA from stdin and print FASTA to stdout, only "
        "including sequences that match the requested taxonomy regular "
        "expression at any level.",
    )

    parser.add_argument(
        "--taxonomy",
        required=True,
        help="The regex to match the taxonomy on. Case is ignored.",
    )

    parser.add_argument(
        "--invert",
        action="store_true",
        default=False,
        help="If True, only write sequences whose taxonomy does not match.",
    )

    parser.add_argument(
        "--detailsFile",
        metavar="FILE",
        default=None,
        help="The name of a file to save taxonomy details to",
    )

    args = parser.parse_args()

    try:
        regexp = re.compile(args.taxonomy, re.I)
    except re.error as e:
        print(
            "Could not compile %r to a regular expression:" % args.taxonomy,
            e,
            file=sys.stderr,
        )
        sys.exit(1)

    detailsFp: Optional[TextIO]

    if args.detailsFile is not None:
        detailsFp = open(args.detailsFile, "w")

        def details(accept, readId, taxonomy):
            return writeDetails(accept, readId, taxonomy, detailsFp)

    else:
        detailsFp = None

        def details(accept, readId, taxonomy):
            return None

    lineageFetcher = LineageFetcher()
    reads = FastaReads(sys.stdin)
    save = sys.stdout.write
    readCount = saveCount = noTaxonomyCount = 0

    for read in reads:
        readCount += 1
        fasta = read.toString("fasta")
        taxonomy = lineageFetcher.lineage(read.id)
        if taxonomy:
            for taxonomyId, scientificName in taxonomy:
                if regexp.match(scientificName):
                    details(True, read.id, taxonomy)
                    if not args.invert:
                        saveCount += 1
                        save(fasta)
                    break
            else:
                details(False, read.id, taxonomy)
                if args.invert:
                    saveCount += 1
                    save(fasta)
        else:
            noTaxonomyCount += 1

    if detailsFp:
        detailsFp.close()

    lineageFetcher.close()

    rejectCount = readCount - saveCount - noTaxonomyCount
    print(
        "%d sequences read, %d (%.2f%%) saved, %d (%.2f%%) rejected, "
        "%d (%.2f%%) no taxomony found."
        % (
            readCount,
            saveCount,
            saveCount / float(readCount) * 100.0,
            rejectCount,
            (rejectCount) / float(readCount) * 100.0,
            noTaxonomyCount,
            noTaxonomyCount / float(readCount) * 100.0,
        ),
        file=sys.stderr,
    )



from unittest import TestCase

from dark.local_align import LocalAlignment
from dark.reads import Read


class TestLocalAlign(TestCase):
    """
    Test the LocalAlignment class.
    With match +1, mismatch -1, gap open -1, gap extend -1 and
        gap extend decay 0.0.
    """

    def testPositiveMismatch(self):
        """
        If the mismatch value passed is positive, an exception
        must be raised.
        """
        seq1 = Read("seq1", "a")
        seq2 = Read("seq2", "a")
        self.assertRaisesRegex(
            ValueError,
            "Mismatch must be negative",
            LocalAlignment,
            seq1,
            seq2,
            mismatch=3,
        )

    def testZeroMismatch(self):
        """
        If the mismatch value passed is zero, an exception
        must be raised.
        """
        seq1 = Read("seq1", "a")
        seq2 = Read("seq2", "a")
        self.assertRaisesRegex(
            ValueError,
            "Mismatch must be negative",
            LocalAlignment,
            seq1,
            seq2,
            mismatch=0,
        )

    def testPositiveGap(self):
        """
        If the gap value passed is positive, an exception
        must be raised.
        """
        seq1 = Read("seq1", "a")
        seq2 = Read("seq2", "a")
        self.assertRaisesRegex(
            ValueError, "Gap must be negative", LocalAlignment, seq1, seq2, gap=3
        )

    def testZeroGap(self):
        """
        If the gap value passed is zero, an exception
        must be raised.
        """
        seq1 = Read("seq1", "a")
        seq2 = Read("seq2", "a")
        self.assertRaisesRegex(
            ValueError, "Gap must be negative", LocalAlignment, seq1, seq2, gap=0
        )

    def testPositiveGapExtend(self):
        """
        If the gap extend value passed is positive, an exception
        must be raised.
        """
        seq1 = Read("seq1", "a")
        seq2 = Read("seq2", "a")
        self.assertRaisesRegex(
            ValueError,
            "Gap extension penalty cannot be positive",
            LocalAlignment,
            seq1,
            seq2,
            gapExtend=3,
        )

    def testFirstSequenceEmpty(self):
        """
        If the first sequence passed is empty, an exception must be raised.
        """
        seq1 = Read("seq1", "")
        seq2 = Read("seq2", "agtcagtcagtc")
        self.assertRaisesRegex(
            ValueError, "Empty sequence: seq1", LocalAlignment, seq1, seq2
        )

    def testSecondSequenceEmpty(self):
        """
        If the second sequence passed is empty, an exception must be raised.
        """
        seq1 = Read("seq1", "agtcagtcagtc")
        seq2 = Read("seq2", "")
        self.assertRaisesRegex(
            ValueError, "Empty sequence: seq2", LocalAlignment, seq1, seq2
        )

    def testBothSequencesEmpty(self):
        """
        If two empty sequences are passed, an exception must be raised.
        """
        seq1 = Read("seq1", "")
        seq2 = Read("seq2", "")
        self.assertRaisesRegex(
            ValueError, "Empty sequence: seq1", LocalAlignment, seq1, seq2
        )

    def testGapAtStartOfSeq1(self):
        seq1 = Read("seq1", "gaatcg")
        seq2 = Read("seq2", "cgaatcg")
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = (
            "\nCigar string of aligned region: 6=\n"
            "seq1 Match start: 1 Match end: 6\n"
            "seq2 Match start: 2 Match end: 7\n"
            "seq1 1 GAATCG 6\n"
            "       ||||||\n"
            "seq2 2 GAATCG 7"
        )
        self.assertEqual(result, alignment)

    def testGapAtStartOfSeq2(self):
        seq1 = Read("seq1", "cgaatcg")
        seq2 = Read("seq2", "gaatcg")
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = (
            "\nCigar string of aligned region: 6=\n"
            "seq1 Match start: 2 Match end: 7\n"
            "seq2 Match start: 1 Match end: 6\n"
            "seq1 2 GAATCG 7\n"
            "       ||||||\n"
            "seq2 1 GAATCG 6"
        )
        self.assertEqual(result, alignment)

    def testGapAtEndOfSeq1(self):
        seq1 = Read("seq1", "cgaatc")
        seq2 = Read("seq2", "cgaatcg")
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = (
            "\nCigar string of aligned region: 6=\n"
            "seq1 Match start: 1 Match end: 6\n"
            "seq2 Match start: 1 Match end: 6\n"
            "seq1 1 CGAATC 6\n"
            "       ||||||\n"
            "seq2 1 CGAATC 6"
        )
        self.assertEqual(result, alignment)

    def testGapAtEndOfSeq2(self):
        seq1 = Read("seq1", "cgaatcg")
        seq2 = Read("seq2", "cgaatc")
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = (
            "\nCigar string of aligned region: 6=\n"
            "seq1 Match start: 1 Match end: 6\n"
            "seq2 Match start: 1 Match end: 6\n"
            "seq1 1 CGAATC 6\n"
            "       ||||||\n"
            "seq2 1 CGAATC 6"
        )
        self.assertEqual(result, alignment)

    def testGapAtBothEndsOfSeq1(self):
        seq1 = Read("seq1", "gaatc")
        seq2 = Read("seq2", "cgaatcg")
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = (
            "\nCigar string of aligned region: 5=\n"
            "seq1 Match start: 1 Match end: 5\n"
            "seq2 Match start: 2 Match end: 6\n"
            "seq1 1 GAATC 5\n"
            "       |||||\n"
            "seq2 2 GAATC 6"
        )
        self.assertEqual(result, alignment)

    def testGapAtBothEndsOfSeq2(self):
        seq1 = Read("seq1", "cgaatcg")
        seq2 = Read("seq2", "gaatc")
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = (
            "\nCigar string of aligned region: 5=\n"
            "seq1 Match start: 2 Match end: 6\n"
            "seq2 Match start: 1 Match end: 5\n"
            "seq1 2 GAATC 6\n"
            "       |||||\n"
            "seq2 1 GAATC 5"
        )
        self.assertEqual(result, alignment)

    def testAlignmentWithGapInMiddle(self):
        seq1 = Read("seq1", "agtcagtcagtc")
        seq2 = Read("seq2", "cgaatcg")
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = (
            "\nCigar string of aligned region: 2=1D1=\n"
            "seq1 Match start: 7 Match end: 10\n"
            "seq2 Match start: 5 Match end: 7\n"
            "seq1 7 TCAG 10\n"
            "       || |\n"
            "seq2 5 TC-G 7"
        )
        self.assertEqual(result, alignment)

    def testTwoEqualSequences(self):
        """
        When two identical sequences are given, the result should
        show that the sequences completely match.
        """
        seq1 = Read("seq1", "cgaatcg")
        seq2 = Read("seq2", "cgaatcg")
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = (
            "\nCigar string of aligned region: 7=\n"
            "seq1 Match start: 1 Match end: 7\n"
            "seq2 Match start: 1 Match end: 7\n"
            "seq1 1 CGAATCG 7\n"
            "       |||||||\n"
            "seq2 1 CGAATCG 7"
        )
        self.assertEqual(result, alignment)

    def testTwoCompletelyDifferentSequences(self):
        """
        When two completely different sequences are given, the result
        should be the two sequences with an empty alignment.
        """
        seq1 = Read("seq1", "aaaaaa")
        seq2 = Read("seq2", "gggggg")
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = "\nNo alignment between seq1 and seq2\n"
        self.assertEqual(result, alignment)

    def testWikiAnswer(self):
        """
        Test the example given in Wikipedia:
        http://en.wikipedia.org/wiki/Smith%E2%80%93Waterman_algorithm
        """
        seq1 = Read("seq1", "ACACACTA")
        seq2 = Read("seq2", "AGCACACA")
        align = LocalAlignment(seq1, seq2, match=2)
        result = align.createAlignment(resultFormat=str)
        alignment = (
            "\nCigar string of aligned region: 1=1I5=1D1=\n"
            "seq1 Match start: 1 Match end: 8\n"
            "seq2 Match start: 1 Match end: 8\n"
            "seq1 1 A-CACACTA 8\n"
            "       | ||||| |\n"
            "seq2 1 AGCACAC-A 8"
        )
        self.assertEqual(result, alignment)

    def testWikiAnswerWithMatchOne(self):
        """
        Test the example given in Wikipedia
        http://en.wikipedia.org/wiki/Smith%E2%80%93Waterman_algorithm
        Wikipedia uses a match score of two, here we use a score of one.
        """
        seq1 = Read("seq1", "ACACACTA")
        seq2 = Read("seq2", "AGCACACA")
        align = LocalAlignment(seq1, seq2, match=1)
        result = align.createAlignment(resultFormat=str)
        alignment = (
            "\nCigar string of aligned region: 5=1D1=\n"
            "seq1 Match start: 2 Match end: 8\n"
            "seq2 Match start: 3 Match end: 8\n"
            "seq1 2 CACACTA 8\n"
            "       ||||| |\n"
            "seq2 3 CACAC-A 8"
        )
        self.assertEqual(result, alignment)

    def testWikiAnswerAsDict(self):
        """
        Test the example given in Wikipedia:
        http://en.wikipedia.org/wiki/Smith%E2%80%93Waterman_algorithm
        with the return result being a dict.
        """
        seq1 = Read("seq1", "ACACACTA")
        seq2 = Read("seq2", "AGCACACA")
        align = LocalAlignment(seq1, seq2, match=2)
        result = align.createAlignment()
        self.assertEqual(
            {
                "cigar": "1=1I5=1D1=",
                "sequence1Start": 1,
                "sequence1End": 8,
                "sequence2Start": 1,
                "sequence2End": 8,
                "text": [
                    "seq1 1 A-CACACTA 8",
                    "       | ||||| |",
                    "seq2 1 AGCACAC-A 8",
                ],
            },
            result,
        )

    def testWikiAnswerWithMatchOneAsDict(self):
        """
        Test the example given in Wikipedia
        http://en.wikipedia.org/wiki/Smith%E2%80%93Waterman_algorithm
        Wikipedia uses a match score of two, here we use a score of one.
        Get the result as a dict.
        """
        seq1 = Read("seq1", "ACACACTA")
        seq2 = Read("seq2", "AGCACACA")
        align = LocalAlignment(seq1, seq2, match=1)
        result = align.createAlignment()
        self.assertEqual(
            {
                "cigar": "5=1D1=",
                "sequence1Start": 2,
                "sequence1End": 8,
                "sequence2Start": 3,
                "sequence2End": 8,
                "text": [
                    "seq1 2 CACACTA 8",
                    "       ||||| |",
                    "seq2 3 CACAC-A 8",
                ],
            },
            result,
        )



#!/usr/bin/env python

import argparse

import requests

# See https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EFetch
# for URL format details.

URL = (
    "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?"
    "db=%s&id=%s&rettype=%s&retmode=text"
)

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=("Fetch a sequence by id from NCBI and write it to stdout as FASTA."),
)

parser.add_argument("id", help="The id of the sequence to fetch.")

parser.add_argument(
    "--database",
    default="nucleotide",
    choices=("nucleotide", "protein"),
    help="The name of the NCBI database to query.",
)

parser.add_argument(
    "--format", default="fasta", choices=("fasta", "gb"), help="The format to return."
)

args = parser.parse_args()

print(requests.get(URL % (args.database, args.id, args.format)).text.rstrip("\n"))



#!/usr/bin/env python

import argparse
import re
import sys

from Bio import SeqIO

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=(
        "Read a GenBank flat file and print GenBank or FASTA for matching records."
    ),
)

parser.add_argument(
    "genbankFile", metavar="FILE.gb", nargs="+", help="The GenBank file(s) to examine."
)

parser.add_argument(
    "--idRegex",
    metavar="REGEX",
    default=".",
    help=(
        "Only print records whose id (or description, if --matchDescription "
        "is used) matches this regular expression."
    ),
)

parser.add_argument(
    "--match",
    choices=("fullmatch", "match", "search"),
    default="search",
    help="How to match the regular expression. See pydoc re for an explanation.",
)

parser.add_argument(
    "--matchDescription",
    action="store_true",
    help="Match on the record description as well as the id.",
)

parser.add_argument(
    "--invert",
    "-v",
    action="store_true",
    help="Invert the match. Only print records that do not match.",
)

parser.add_argument(
    "--ignoreCase", "-i", action="store_true", help="Use case-insensitive matching."
)

parser.add_argument(
    "--format", choices=("fasta", "gb"), default="gb", help="The output format."
)

parser.add_argument(
    "--expectedCount",
    type=int,
    help=(
        "The expected number of records to find. If this number is not "
        "found, print a message and exit non-zero."
    ),
)

parser.add_argument(
    "--verbose",
    action="store_true",
    help="Print a summary of records read and printed.",
)

args = parser.parse_args()

recordCount = matchCount = 0
match = getattr(
    re.compile(args.idRegex, flags=(re.I if args.ignoreCase else 0)), args.match
)

matchDescription = args.matchDescription
invert = args.invert


def filterFunc(records):
    global matchCount, recordCount
    for record in records:
        recordCount += 1
        id_ = record.id + (f" {record.description}" if record.description else "")
        if match(id_ if matchDescription else record.id):
            emit = not invert
        else:
            emit = invert

        if emit:
            matchCount += 1
            yield record


for filename in args.genbankFile:
    with open(filename) as fp:
        SeqIO.write(filterFunc(SeqIO.parse(fp, "gb")), sys.stdout, args.format)

if args.verbose:
    print(
        f"Read {recordCount} record{'s' if recordCount > 1 else ''}, "
        f"printed {matchCount}",
        file=sys.stderr,
    )

if args.expectedCount is not None and matchCount != args.expectedCount:
    print(
        f"Expected {args.expectedCount} records but found {matchCount}. Exiting.",
        file=sys.stderr,
    )
    sys.exit(1)



#!/usr/bin/env python

"""
Read a file of our JSON BLAST output (which has one JSON object per
line) from stdin and pretty print it to stdout.
"""

import sys
from json import dumps, loads

for line in sys.stdin:
    s = dumps(loads(line[:-1]), sort_keys=True, indent=2)
    print("\n".join([field.rstrip() for field in s.splitlines()]))



#!/usr/bin/env python

import argparse
import sys
from os import mkdir
from os.path import exists, join
from tempfile import mkdtemp

from dark.process import Executor


def makeOuputDir(outputDir, force):
    """
    Create or check for an output directory.

    @param outputDir: A C{str} output directory name, or C{None}.
    @param force: If C{True}, allow overwriting of pre-existing files.
    @return: The C{str} output directory name.
    """
    if outputDir:
        if exists(outputDir):
            if not force:
                print(
                    "Will not overwrite pre-existing files. Use --force to make me.",
                    file=sys.stderr,
                )
                sys.exit(1)
        else:
            mkdir(outputDir)
    else:
        outputDir = mkdtemp()
        print("Writing output files to %s" % outputDir)

    return outputDir


def samtoolsMpileup(outFile, referenceFile, alignmentFile, executor):
    """
    Use samtools mpileup to generate VCF.

    @param outFile: The C{str} name to write the output to.
    @param referenceFile: The C{str} name of the FASTA file with the reference
        sequence.
    @param alignmentFile: The C{str} name of the SAM or BAM alignment file.
    @param executor: An C{Executor} instance.
    """
    executor.execute(
        "samtools mpileup -u -v -f %s %s > %s" % (referenceFile, alignmentFile, outFile)
    )


def bcftoolsMpileup(outFile, referenceFile, alignmentFile, executor):
    """
    Use bcftools mpileup to generate VCF.

    @param outFile: The C{str} name to write the output to.
    @param referenceFile: The C{str} name of the FASTA file with the reference
        sequence.
    @param alignmentFile: The C{str} name of the SAM or BAM alignment file.
    @param executor: An C{Executor} instance.
    """
    executor.execute(
        "bcftools mpileup -Ov -f %s %s > %s" % (referenceFile, alignmentFile, outFile)
    )


def bcftoolsCallMulti(outFile, vcfFile, executor):
    """
    Use bcftools to make consensus calls, using the  -m (multiallelic) option.

    @param outFile: The C{str} name to write the output to.
    @param vcfFile: The C{str} name of the VCF file with the calls from
        the pileup.
    @param executor: An C{Executor} instance.
    """
    # Note that this just gives a call on the first genome in the calls file.
    executor.execute("bcftools call -m -Ov -o %s < %s" % (outFile, vcfFile))


def bcftoolsCallConsensus(outFile, vcfFile, executor):
    """
    Use bcftools to make consensus calls, using the original -c option.

    @param outFile: The C{str} name to write the output to.
    @param vcfFile: The C{str} name of the VCF file with the calls from
        the pileup.
    @param executor: An C{Executor} instance.
    """
    # Note that this just gives a call on the first genome in the calls file.
    executor.execute("bcftools call -c -Ov -o %s < %s" % (outFile, vcfFile))


def bcftoolsConsensus(outFile, vcfFile, id_, referenceFile, executor):
    """
    Use bcftools to extract consensus FASTA.

    @param outFile: The C{str} name to write the output to.
    @param vcfFile: The C{str} name of the VCF file with the calls from
        the pileup.
    @param id_: The C{str} identifier to use in the resulting FASTA sequence.
    @param referenceFile: The C{str} name of the FASTA file with the reference
        sequence.
    @param executor: An C{Executor} instance.
    """
    bgz = vcfFile + ".gz"
    executor.execute("bgzip -c %s > %s" % (vcfFile, bgz))
    executor.execute("tabix %s" % bgz)
    executor.execute(
        "bcftools consensus %s < %s | "
        "filter-fasta.py --idLambda 'lambda id: \"%s\"' > %s"
        % (bgz, referenceFile, id_, outFile)
    )


def vcfutilsConsensus(outFile, vcfFile, id_, _, executor):
    """
    Use vcftools to extract consensus FASTA.

    @param outFile: The C{str} name to write the output to.
    @param vcfFile: The C{str} name of the VCF file with the calls from
        the pileup.
    @param id_: The C{str} identifier to use in the resulting FASTA sequence.
    @param executor: An C{Executor} instance.
    """
    executor.execute(
        "vcfutils.pl vcf2fq < %s | "
        "filter-fasta.py --fastq --quiet --saveAs fasta "
        "--idLambda 'lambda id: \"%s\"' > %s" % (vcfFile, id_, outFile)
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Compare different consensus making methods.",
    )

    parser.add_argument(
        "--referenceFile",
        required=True,
        metavar="FILENAME",
        help="The name of the FASTA file containing the reference sequence.",
    )

    parser.add_argument(
        "--alignmentFile",
        required=True,
        metavar="FILENAME",
        help=(
            "The name of the SAM or BAM file containing an alignment to the reference."
        ),
    )

    parser.add_argument(
        "--verbose",
        type=int,
        default=0,
        metavar="N",
        help=(
            "The integer verbosity level (0 = no output, 1 = some output, "
            "2 = maximal output)."
        ),
    )

    parser.add_argument(
        "--force",
        default=False,
        action="store_true",
        help="If given, overwrite pre-existing files.",
    )

    parser.add_argument(
        "--outputDir", metavar="DIRNAME", help="The directory to save result files to."
    )

    args = parser.parse_args()

    outputDir = makeOuputDir(args.outputDir, args.force)

    executor = Executor()

    pileuppers = (
        ("samtools-mpileup", samtoolsMpileup),
        ("bcftools-mpileup", bcftoolsMpileup),
    )

    callers = (("bcftools-c", bcftoolsCallConsensus), ("bcftools-m", bcftoolsCallMulti))

    consensusers = (
        ("bcftools-consensus", bcftoolsConsensus),
        ("vcfutils-vcf2fq", vcfutilsConsensus),
    )

    consensusFiles = []

    for pileupName, pileupFunc in pileuppers:
        pileupMiddle = pileupName
        pileupFile = join(outputDir, "pileup-" + pileupMiddle + ".vcf")
        pileupFunc(pileupFile, args.referenceFile, args.alignmentFile, executor)

        for callerName, callerFunc in callers:
            callerMiddle = pileupMiddle + "-" + callerName
            callerFile = join(outputDir, "calls-" + callerMiddle + ".vcf")
            callerFunc(callerFile, pileupFile, executor)

            for consensusName, consensusFunc in consensusers:
                consensusMiddle = callerMiddle + "-" + consensusName
                consensusFile = join(
                    outputDir, "consensus-" + consensusMiddle + ".fasta"
                )
                consensusFiles.append(consensusFile)
                consensusFunc(
                    consensusFile,
                    callerFile,
                    consensusMiddle,
                    args.referenceFile,
                    executor,
                )

    # Let's assume there's at least one consensus file.
    consensusesFile = join(outputDir, "consensuses.fasta")
    executor.execute("cat %s > %s" % (" ".join(consensusFiles), consensusesFile))

    htmlFile = join(outputDir, "consensus-identity.html")
    executor.execute(
        (
            "fasta-identity-table.py --footer --showGaps --showLengths < %s | "
            "perl -pe 's/-(bcftools|vcfutils)/ $1/g' > %s"
        )
        % (consensusesFile, htmlFile)
    )

    verbose = args.verbose
    if verbose > 0:
        print("The following commands were executed:")
        for line in executor.log:
            if line.startswith("#"):
                if verbose > 1:
                    print(line)
            else:
                print(line)

    print(
        "Identity table comparing %d consensuses written to %s"
        % (len(consensusFiles), htmlFile)
    )



import sys
from pathlib import Path

from dark.fasta import FastaReads
from dark.process import Executor


def extract_fasta(
    fasta: Path,
    ref_id: str,
    out_dir: Path,
    index_fasta: bool,
    quiet: bool,
    force: bool,
    keep_description: bool,
    e: Executor,
) -> int:
    new_reference = out_dir / f"{ref_id}.fasta"

    if new_reference.exists() and not force:
        sys.exit(
            f"Output FASTA file {str(new_reference)!r} already exists. Exiting. "
            "Re-run with --force to overwrite it."
        )

    # Find the reference in the FASTA and write it to the new_reference file.  In
    # writing it, make sure it has the same id as the reference we're going to look for
    # in the BAM file. We drop anything after the first space in the ID in the FASTA
    # file because that cannot appear in the BAM/SAM file we'll search (and write).
    for read in FastaReads(fasta):
        found = truncated = False
        if read.id == ref_id:
            found = True
            truncated = False
        elif not keep_description and read.id.split()[0] == ref_id:
            found = truncated = True

        if found:
            if truncated:
                if not quiet and not keep_description:
                    print(
                        f"Reference ID {ref_id!r} was found in {str(fasta)!r} but with "
                        f"a description ({read.id!r}). The description will be dropped "
                        "in the output FASTA so that it matches the reference name in "
                        "the output BAM (since spaces in an ID in a BAM file are "
                        "illegal). This is done so the IDs in both files match, as "
                        "required by some tools. To instead keep the description in "
                        "the output FASTA, re-run with --keep-description."
                    )
                read.id = ref_id

            with open(new_reference, "w") as fp:
                print(read.toString(), end="", file=fp)

            if index_fasta:
                e.execute(f"samtools faidx {str(new_reference)!r}")

            return len(read)

    if keep_description:
        # Have another look, dropping the description. Even though we were told not to,
        # pointing out that the ID is found but with a description is likely to be
        # helpful.
        #
        # Note that this relies on being able to re-read the FASTA. I.e., it must be a
        # file not an already open file descriptor (not an iterator that gets exhausted
        # on first use).
        for read in FastaReads(fasta):
            if read.id.split()[0] == ref_id:
                sys.exit(
                    f"A sequence with ID {ref_id!r} (exactly) does not appear in "
                    f"{str(fasta)!r}. However, that FASTA file has at least one "
                    f"sequence with an ID ({read.id!r}) whose first part matches your "
                    "reference ID. But, you specified --keep-description, so this ID "
                    "with its description has not been considered as matching. Maybe "
                    "try re-running without --keep-description?"
                )

    sys.exit(f"Could not find reference {ref_id!r} in {str(fasta)!r}.")


def extract_sam(
    bam: Path,
    ref_id: str,
    reference_len: int,
    out_dir: Path,
    keep_sam: bool,
    force: bool,
    e: Executor,
) -> int:
    new_sam = out_dir / f"{ref_id}.sam"
    new_bam = new_sam.with_suffix(".bam")

    for f in new_sam, new_bam:
        if f.exists() and not force:
            sys.exit(
                f"Output file {str(f)!r} already exists. Exiting. "
                "Re-run with --force to overwrite it."
            )

    if bam.name.lower().endswith(".bam"):
        # Convert the given BAM to SAM so we can easily read it (we could instead use
        # the pysam module).
        sam = out_dir / bam.with_suffix(".sam").name
        e.execute(f"samtools view -h -O SAM {str(bam)!r} --output {str(sam)!r}")
        our_sam = True
    else:
        if not bam.name.lower().endswith(".sam"):
            sys.exit(
                f"Input BAM/SAM file {str(bam)!r} does not have a .bam or .sam "
                "suffix so I don't know how to read it."
            )
        sam = bam
        our_sam = False

    read_count = 0

    with open(sam) as sam_in, open(new_sam, "w") as sam_out:
        for line in sam_in:
            line = line.strip()
            fields = line.split("\t", maxsplit=3)
            if line.startswith("@"):
                if fields[0] == "@SQ":
                    if fields[1] == f"SN:{ref_id}":
                        if fields[2] == f"LN:{reference_len}":
                            print(line, file=sam_out)
                        else:
                            sys.exit(
                                f"Found reference {ref_id!r} in "
                                f"{str(sam)!r} but the length field in the SAM "
                                f"file is {fields[2]!r} whereas the length of the "
                                f"reference in {str(ref_id)!r} is "
                                f"{reference_len}."
                            )
                else:
                    # A non-@SQ header line.
                    print(line, file=sam_out)
            else:
                # A non-header line. Print it if this is a match on the reference.
                if fields[2] == ref_id:
                    read_count += 1
                    print(line, file=sam_out)

    e.execute(f"samtools sort -O BAM {str(new_sam)!r} -o {str(new_bam)!r}")
    e.execute(f"samtools index {str(new_bam)!r}")

    # Remove the intermediate SAM, if we made it. This was just a direct conversion of
    # the input BAM that we made so we could read it easily to find the matching records
    # (which we wrote to new_sam).
    if our_sam:
        sam.unlink()

    if not keep_sam:
        new_sam.unlink()

    return read_count



#!/usr/bin/env python

import argparse
import sys

from dark.reads import DNARead, addFASTACommandLineOptions, parseFASTACommandLineOptions

FORWARD, REVERSE, COMPLEMENT, REVERSE_COMPLEMENT = 0, 1, 2, 3

SUFFIX = {
    FORWARD: "",
    REVERSE: " (r)",
    COMPLEMENT: " (c)",
    REVERSE_COMPLEMENT: " (rc)",
}


def getParser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=(
            "Given FASTA on stdin and a sequence (or genome region) to find, write "
            "information about (potentially overlapping) matches (and, optionally, "
            "reversed, complemented, or reversed complement matches), to standard "
            "output. Offsets are inclusive and 1-based unless --zeroBased is used "
            "(in which case they are 0-based closed/open offsets as used in Python). "
        )
    )

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument("--pattern", "--sequence", help="The sequence to look for.")

    group.add_argument(
        "--region",
        help=(
            "A region of the form A-B, where A and B are numeric sites, with "
            "A <= B. A and B will be treated as 1-based offsets unless "
            "--zeroBased is used. This can be used to find repeat copies of an "
            "input sequence region. I.e., for each input sequence, the specified "
            "region of the sequence will be searched for in the rest of the sequence."
        ),
    )

    parser.add_argument(
        "--reverseComplement",
        "--rc",
        action="store_true",
        help=(
            "Also search for the reverse complement. Reverse complement matches "
            f"are indicated with an{SUFFIX[REVERSE_COMPLEMENT]} suffix."
        ),
    )

    parser.add_argument(
        "--reverse",
        action="store_true",
        help=(
            f"Also search for the reversed sequence. Reverse matches are "
            f"indicated with an{SUFFIX[REVERSE]} suffix."
        ),
    )

    parser.add_argument(
        "--complement",
        action="store_true",
        help=(
            f"Also search for the complemented sequence. Complement matches "
            f"are indicated with a{SUFFIX[COMPLEMENT]} suffix."
        ),
    )

    parser.add_argument(
        "--end",
        action="store_true",
        help="Print the index of the end of the match, not the start.",
    )

    parser.add_argument(
        "--offsetsOnly",
        action="store_true",
        help=(
            "Only print the matching offset (or offsets, separated by the --sep value)."
        ),
    )

    parser.add_argument(
        "--checkMatchCount",
        type=int,
        help=(
            "The expected match count. This can be used to force an error if the "
            "number of occurrences of the search pattern found is not as expected."
        ),
    )

    parser.add_argument(
        "--caseSensitive",
        action="store_true",
        help="Do not ignore upper/lower case when matching.",
    )

    parser.add_argument(
        "--printSequences",
        action="store_true",
        help=(
            f"Print the sequences searched for. Reversed, complemented, and "
            f"reverse-complemented sequences are indicated by "
            f"{SUFFIX[REVERSE]},{SUFFIX[COMPLEMENT]}, "
            f"and{SUFFIX[REVERSE_COMPLEMENT]}, respectively."
        ),
    )

    parser.add_argument(
        "--printTrivialMatch",
        action="store_true",
        help=(
            "If --region is used, there is a guaranteed match at that "
            "location. Do not include that match in the printed matches "
            "unless this option is given."
        ),
    )

    parser.add_argument(
        "--sep", default="\t", help="The output field separator string (default=TAB)."
    )

    parser.add_argument(
        "--zeroBased",
        action="store_true",
        help=(
            "Return Python-style zero-based closed/open match offsets. When --end is "
            "used, the output offset will be for the character that follows the end of "
            "the match (i.e., it is not part of the match)."
        ),
    )

    parser.add_argument(
        "--gapCharacter",
        default="-",
        help="The gap character (only used if --ignoreGaps is used).",
    )

    parser.add_argument(
        "--ignoreGaps",
        action="store_true",
        help=(
            "Ignore gaps when matching. This is useful for finding a subsequence of "
            "an aligned sequence. Note that the reported index into input sequences "
            "will be for the (possibly gapped) original input sequence."
        ),
    )

    addFASTACommandLineOptions(parser)

    return parser


def parseRegion(region: str, zeroBased: bool) -> tuple[int, int]:
    """
    Parse a region string like 34-79 and return the two numbers.
    """
    lo, hi = map(int, region.split("-"))

    if zeroBased:
        if not (0 <= lo <= hi):
            sys.exit(
                f"Invalid range {region!r}. Give two non-negative numbers "
                "in non-decreasing order.",
            )
    else:
        if not (0 < lo <= hi):
            extra = (
                " Or use --zeroBased if you really want to specify a zero offset."
                if lo == 0
                else ""
            )

            sys.exit(
                f"Invalid range {region!r}. Give two positive numbers in "
                f"non-decreasing order.{extra}",
            )

        # Use zero-based indices.
        lo -= 1

    return lo, hi


def getPatterns(
    pattern: str, reverse: bool, complement: bool, reverseComplement: bool
) -> list[tuple[str, int]]:
    """
    Get a list of patterns (and their orientations) to search for.
    """
    # We always look for the given pattern in the forward direction.
    result = [(pattern, FORWARD)]

    if reverse:
        result.append((pattern[::-1], REVERSE))

    rc = DNARead("id", pattern).reverseComplement().sequence

    if reverseComplement:
        result.append((rc, REVERSE_COMPLEMENT))

    if complement:
        # The complement is just the reverse of the reverse complement.
        result.append((rc[::-1], COMPLEMENT))

    return result


def main() -> None:
    args = getParser().parse_args()

    if args.region:
        lo, hi = parseRegion(args.region, args.zeroBased)
        patterns = []  # Will be set below, extracted from each target read.
    else:
        lo = hi = -1  # Unused.
        patterns = getPatterns(
            args.pattern, args.reverse, args.complement, args.reverseComplement
        )

    reads = parseFASTACommandLineOptions(args)

    for read in reads:
        if args.region:
            patterns = getPatterns(
                read.sequence[lo:hi],
                args.reverse,
                args.complement,
                args.reverseComplement,
            )

        matches: set[tuple[int, int]] = set()

        for pattern, orientation in patterns:
            # Look for as many occurrences of this pattern as can be found in the read
            # sequence.
            start = 0
            while True:
                index = read.find(
                    pattern,
                    start=start,
                    end=args.end,
                    caseSensitive=args.caseSensitive,
                    ignoreGaps=args.ignoreGaps,
                    gapCharacter=args.gapCharacter,
                )
                if index == -1:
                    # No more matches of this pattern.
                    break
                else:
                    if args.region is None or args.printTrivialMatch or index != lo:
                        assert (index, orientation) not in matches
                        matches.add((index, orientation))
                    start = index + 1

        if args.checkMatchCount is not None and len(matches) != args.checkMatchCount:
            sys.exit(
                f"{len(matches)} {'match was' if len(matches) == 1 else 'matches were'} "  # noqa: E501
                f"found for sequence {read.id!r}, but {args.checkMatchCount} "
                f"{'match was' if args.checkMatchCount == 1 else 'matches were'} "
                "expected."
            )

        if matches:
            adjust = 0 if args.end or args.zeroBased else 1

            if args.offsetsOnly:
                fields = [str(index + adjust) for index, _ in sorted(matches)]
            else:
                fields = [
                    read.id,
                    f"{len(matches)} match{'' if len(matches) == 1 else 'es'}",
                    ", ".join(
                        str(index + adjust) + SUFFIX[orientation]
                        for index, orientation in sorted(matches)
                    ),
                ]

                if args.printSequences:
                    for pattern, orientation in patterns:
                        fields.append(f"sequence{SUFFIX[orientation]} = {pattern}")

            print(args.sep.join(fields))


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import sys

from rich import print as rprint

from dark.aa import AminoAcid, find
from dark.aaVars import CODONS, PROPERTY_NAMES, STOP_CODONS
from dark.codonDistance import codonInformation

# Nucleotide colors. Note that 'A' will just be rendered as bold text in
# whatever the default terminal color is (often black or white, depending on
# the user's preferences). This is not perfect, but it's hard to be perfect.
COLORS = {
    "A": "",
    "C": "red",
    "G": "green",
    "T": "blue",
}


def colorCodons(codon1: str, codon2: str) -> tuple[str, str]:
    """
    Take two codons and return the same two codons but with differing
    nucleotides colored.
    """
    result1 = result2 = ""

    for nt1, nt2 in zip(codon1, codon2):
        if nt1 == nt2:
            result1 += nt1
            result2 += nt2
        else:
            color1 = f"bold {COLORS[nt1]}"
            color2 = f"bold {COLORS[nt2]}"
            result1 += f"[{color1}]{nt1}[/{color1}]"
            result2 += f"[{color2}]{nt2}[/{color2}]"

    return result1, result2


def findOrDie(s) -> AminoAcid:
    """
    Look up an amino acid.

    @param s: A C{str} amino acid specifier. This may be a full name,
        a 3-letter abbreviation or a 1-letter abbreviation. Case is ignored.
    @return: An C{AminoAcid} instance, if one can be found. Else exit.
    """
    aas = list(find(s))
    if len(aas) == 1:
        return aas[0]

    if aas:
        print(
            f"Multiple amino acids matched {s!r}:",
            ", ".join(sorted(aa.name for aa in aas)),
            file=sys.stderr,
        )
    else:
        if s.upper() in STOP_CODONS:
            print(f"Argument {s!r} is a stop codon!", file=sys.stderr)
        else:
            print(f"Unknown amino acid or codon: {s!r}", file=sys.stderr)
        print(
            "Valid amino acid arguments are: %s." % ", ".join(sorted(CODONS)),
            file=sys.stderr,
        )

    sys.exit(1)


def main():
    def codonKey(codonInfo):
        return codonInfo[0], codonInfo[1]

    def transitionsKey(codonInfo):
        return codonInfo[2]

    if len(sys.argv) != 3:
        print("Usage: %s amino_acid amino_acid" % sys.argv[0], file=sys.stderr)
        sys.exit(1)

    arg1 = sys.argv[1]
    arg2 = sys.argv[2]

    if arg1 == arg2:
        print("Identical arguments, nothing to do.", file=sys.stderr)
        sys.exit(1)

    aa1 = findOrDie(arg1)
    aa2 = findOrDie(arg2)

    if aa1.name == aa2.name:
        print(
            f"Arguments {arg1!r} and {arg2!r} both correspond to {aa1.name}, "
            "nothing to do.",
            file=sys.stderr,
        )
        sys.exit(1)

    print("Names:")
    print(f"  {arg1}: {aa1.name}")
    print(f"  {arg2}: {aa2.name}")

    def propertyNames(properties: set[int]) -> str:
        return ", ".join(PROPERTY_NAMES[p] for p in properties)

    aa1Properties = aa1.propertySet()
    aa2Properties = aa2.propertySet()

    print("Properties:")
    print(f"  {aa1.name}: {propertyNames(aa1Properties)}")
    print(f"  {aa2.name}: {propertyNames(aa2Properties)}")

    # aa1Only = propertyNames(aa1Properties - aa2Properties)
    # if aa1Only:
    #     print(f"  {aa1.name} only: {aa1Only}")

    # aa2Only = propertyNames(aa2Properties - aa1Properties)
    # if aa2Only:
    #     print(f"  {aa2.name} only: {aa2Only}")

    common = propertyNames(aa1Properties & aa2Properties) or "Nothing"
    print(f"  In common: {common}")

    print("Codons:")
    print("  %s: %s" % (aa1.name, ", ".join(sorted(aa1.codons))))
    print("  %s: %s" % (aa2.name, ", ".join(sorted(aa2.codons))))

    # Codon distance output is sorted first by distance, then by number of
    # transitions (lowest to highest) then by codon DNA. The idea being to
    # present the possible codon changes to get from one amino acid to
    # another in the order that requires the least change to the most.

    print("Codon distances:")
    distances = codonInformation(aa1.codons, aa2.codons, countTransitions=True)

    for distance in sorted(distances):
        for codon1, codon2, transitions in sorted(
            sorted(distances[distance], key=codonKey), key=transitionsKey, reverse=True
        ):
            transversions = distance - transitions
            codon1, codon2 = colorCodons(codon1, codon2)
            rprint(
                f"  {distance}: {aa1.name}: {codon1} -> {aa2.name}: {codon2}",
                end=" ",
            )
            print(f"(transitions: {transitions}, transversions {transversions})")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import hashlib
import sys

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            "Given FASTA on stdin, write it to stdout according to a given format."
        )
    )

    parser.add_argument(
        "--prefix",
        help=(
            "The string prefix to be used. E.g., use --prefix r for raw "
            'strings. The default is "f" (i.e., use Python f-strings) for '
            'Python 3.6 and later, or else "" (i.e., regular strings).'
        ),
    )

    parser.add_argument(
        "--start",
        default=1,
        type=int,
        help='The integer to start counting from, for the "count" variable.',
    )

    parser.add_argument(
        "--end",
        help=(
            'A string to use as the "end" keyword argument for print(). '
            "This will be passed to eval, allowing convenient use of "
            "Python string constructs such as \\t on the command line. "
            "If not specified, a newline will be printed."
        ),
    )

    parser.add_argument(
        "--format",
        help=(
            "The output format. The default is to reproduce the input. "
            "The format can be that used by Python to specify items in a "
            'dictionary. Available dictionary items are "id", "sequence", '
            '"quality", "length", "count", and "md5" (of the sequence). A '
            "dark.reads.Read instance (named 'read') and an 'md5' "
            "function (that takes and returns a string) are both also in "
            "scope for each FASTA record and these can be accessed using "
            "f-string format (with Python 3.6 or greater). The passed "
            "format string will be eval'd in a double quoted string, as "
            "with the --end string, allowing the use of Python string "
            "conveniences such as \\t on the command line."
        ),
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    format_ = args.format
    end = "\n" if args.end is None else eval('"' + args.end + '"')

    if args.prefix is None:
        prefix = "" if sys.version_info < (3, 6) else "f"
    else:
        prefix = args.prefix

    if format_ is None:
        format_ = (
            r">%(id)s\n%(sequence)s"
            if args.fasta
            else r"@%(id)s\n%(sequence)s\n+\n%(quality)s"
        )

    needMd5 = "%(md5)" in format_
    needLength = "%(length)" in format_

    def md5(s):
        return hashlib.md5(s.encode("utf-8")).hexdigest()

    for count, read in enumerate(reads, start=args.start):
        sequence = read.sequence
        d = {
            "count": count,
            "id": read.id,
            "quality": read.quality,
            "sequence": sequence,
        }
        if needMd5:
            d["md5"] = md5(sequence)
        if needLength:
            d["length"] = len(sequence)

        print(eval(prefix + '"' + format_ + '" % d'), end=end)



import sys
import tempfile
from typing import Iterable, TextIO

import dendropy
from ete4 import Tree


def removeAlrt(
    treeFile: str,
    supportThreshold: int,
    alrtThreshold: float,
    adjustSupportValues: bool,
    replacementSupportValue: int | str,
    expectedUnlabeledNodeCount: int,
    verbose: bool = False,
    verboseFp: TextIO = sys.stderr,
) -> str:
    """
    Read a Newick tree file produced by iqtree2 with both alrt and support
    values on nodes, drop the alrt values and optionally adjust the support
    values (based on the passed bootstrap support threshold and alrt
    threshold). Return the new tree as a (Newick) string.

    This function was written when using baltic
    (https://github.com/evogytis/baltic/tree/master) to draw trees.  Baltic
    cannot deal with node labels like 83.7/95 that are produced by iqtree2
    when you ask it for both bootstrap and alrt values (see
    http://www.iqtree.org/doc/Frequently-Asked-Questions). ete also cannot
    read those labels. But dendropy can. So we read the tree file with
    dendropy, examine the bootstrap support and alrt values, and return a
    tree with just (opionally adjusted) bootstrap support values.

    If C{adjustSupportValues} is True, the support on branches that have
    support values at least C{supportThreshold} but alrt values below
    C{alrtThreshold} will be set to the replacement support value.

    This allows you to put an artificially low support value onto such
    branches so that in display of the tree you can put a '*' onto branches
    with high support and nothing onto the ones with support below the
    threshold. I.e., this is a hack to take both support and alrt values into
    account in a context where you just want to display either a '*' or nothing
    at all on branches (the branches with the replacement support value will not
    have anything shown). Hopefully that makes sense!

    @param treeFile: The C{str} name of the Newick tree file to read.
    @param supportThreshold: The C{int} minimum support value needed to qualify for
        as high-support when C{adjustSupportValues} is true.
    @param alrtThreshold: The C{float} minimum alrt value needed to qualify for
        as high-support when C{adjustSupportValues} is true.
    @param adjustSupportValues: If C{True}, adjust the support on branches that have
        support values at least C{supportThreshold} but alrt values below
        C{alrtThreshold} to have a support value that is the passed support threshold
        minus one.
    @param replacementSupportValue: The C{int} or C{str} value to use as a label
        for branch whose support / alrt values are overall too low.
    @param expectedUnlabeledNodeCount: The number of unlabeled (internal) nodes
        expected. Pass -1 to allow any number of unlabeled internal nodes.
    @param verbose: If true, print information about nodes with good support but low
        alrt values to the C{verboseFp} output stream.
    @param verboseFp: The open output stream to write to if C{verbose} is true.

    """
    replacementSupportLabel = str(
        supportThreshold - 1
        if replacementSupportValue is None
        else replacementSupportValue
    )

    with open(treeFile) as fp:
        tree = dendropy.Tree.get(file=fp, schema="newick", preserve_underscores=True)

    unsupportedDueToLowAlrt = unlabeledNodeCount = nodeCount = 0

    for nodeCount, node in enumerate(
        tree.preorder_internal_node_iter(exclude_seed_node=False), start=1
    ):
        try:
            alrts, supports = node.label.split("/")
        except ValueError:
            raise ValueError(
                f"WARNING: node value ({node.label!r}) on node {node!r} could "
                "not be split on '/'.",
            )
        except AttributeError:
            # This can happen because the tree has been rooted and one edge
            # leading away from the inserted root node leads to a node with
            # no support label (this always happens when rooting on an edge
            # that leads to a tip). Or because a root node was created with
            # no label.
            assert node.label is None
            unlabeledNodeCount += 1
            continue
        else:
            node.label = supports
            if adjustSupportValues:
                support = int(supports)
                alrt = float(alrts)
                if support >= supportThreshold and alrt < alrtThreshold:
                    # When drawn, the edge coming to this node should not
                    # have a star, because its combined support / alrt values
                    # aren't high enough.  So we set its support label to the
                    # special value.
                    node.label = replacementSupportLabel
                    unsupportedDueToLowAlrt += 1

    if (
        expectedUnlabeledNodeCount != -1
        and unlabeledNodeCount != expectedUnlabeledNodeCount
    ):
        raise ValueError(
            f"Found {unlabeledNodeCount} internal nodes with no label "
            f"in {treeFile!r}, but {expectedUnlabeledNodeCount} were expected."
        )

    # Sanity check that there are now no node labels with a slash in them.
    for node in tree.preorder_internal_node_iter(exclude_seed_node=True):
        assert node.label is None or "/" not in node.label

    if verbose:
        print(
            f"Of {nodeCount} internal nodes, {unsupportedDueToLowAlrt} had "
            f"an alrt value that did not meet the {alrtThreshold} threshold, "
            "despite having a sufficiently high bootstrap support "
            f"({supportThreshold}) to otherwise qualify as well-supported.",
            file=verboseFp,
        )

    with tempfile.NamedTemporaryFile() as new:
        tree.write(path=new.name, schema="newick")
        result = open(new.name).read().strip()

    return result


def ete4root(
    tree: Tree, tipNames: Iterable[str], detach: bool = False, scale: float = 1.0
) -> Tree:
    """
    Root an ete tree at the branch leading into the MRCA of the given node
    (names), and then optionally detach that node.

    @param tree: An ete C{Tree} instance.
    @param tipNames: A list of tip names. The tree will be re-rooted on the branch
        leading into the MRCA of these tips.
    @param detach: After re-rooting, detach the outgroup.
    @param scale: The distances from the new root to its two children will normally
        both be half the distance on the branch where the new root is inserted
        (i.e., the branch leading into the MRCA of C{tipNames}). If you pass a
        scale value, the distance leading to the two children will instead be half
        that branch's distance divided by the scale factor. This can be useful for
        displaying a tree where the
    """
    tips = []

    for tipName in tipNames:
        nodes = list(tree.search_nodes(name=tipName))
        if len(nodes) == 0:
            sys.exit(f"Could not find a tip node with name matching {tipName!r}.")
        elif len(nodes) > 1:
            names = ", ".join(sorted(node.name for node in nodes))
            sys.exit(
                f"Found {len(nodes)} tip nodes with names matching {tipName!r}. "
                f"Found node names: {names}."
            )
        else:
            tips.append(nodes[0])

    if len(tips) == 1:
        mrca = tips[0]
    else:
        mrca = tree.common_ancestor(tips)

        # Special case the situation where we have an unrooted tree (i.e.,
        # with a root that has three children) and the nodes on the paths
        # back to the MRCA of all the wanted tips include two of the
        # children. Then we can re-root on the branch leading into the child
        # that was not included. This results in a new root that has that
        # excluded child as one child and the second child is the former
        # root, which (due to the insertion of the new root node) now has
        # just two children (the two that are the MRCA of the set of tips).
        if len(mrca.children) == 3:
            assert mrca.is_root
            included = set(tips)
            for tip in tips:
                ancestors = tip.ancestors(include_root=False)
                included.update(ancestors)
            excluded = set(mrca.children) - included
            if len(excluded) == 1:
                mrca = excluded.pop()
                assert not mrca.is_root
        elif mrca.is_root:
            # We cannot re-root on the branch leading to the MRCA because the
            # MRCA is the root.  If we called tree.set_outgroup(mrca)
            # we would get an ete TreeError: 'Cannot set myself as outgroup'.
            print(
                "The MRCA of the given tips is already the tree root. Returning "
                "without re-rooting.",
                file=sys.stderr,
            )
            return tree

    # Setting the MRCA node as the outgroup creates a new root node, which we can then
    # access via mrca.up
    tree.set_outgroup(mrca)
    root = mrca.up
    assert root.is_root and len(root.children) == 2

    if scale != 1.0:
        assert scale > 0.0, "--scale must be greater than zero."
        for node in root.children:
            if node.dist is not None:
                node.dist /= scale

    if detach:
        mrca.detach()
        assert len(root.children) == 1, (
            f"Root has {len(root.children)} children (expected 1)."
        )
        tree = root.children[0]
        tree.dist = 0.0
        tree.up = None

    return tree


# Backwards compatibility.
ete3root = ete4root



import bz2
from collections import Counter
from json import dumps, loads
from operator import itemgetter

from Bio.File import as_handle

from dark.alignments import Alignment, ReadAlignments
from dark.diamond.hsp import normalizeHSP
from dark.hsp import HSP, LSP
from dark.score import HigherIsBetterScore

# The following are the fields (in the order they are expected on the
# command line via --outfmt 6) that must be given to DIAMOND blastx to
# allow its output to be parsed by convert-diamond-to-sam.py (which uses
# diamondTabularFormatToDicts (below)).
FIELDS = (
    "bitscore btop qframe qend full_qqual qlen full_qseq qseqid "
    "qstart slen sstart stitle"
)

# The keys in the following are DIAMOND format 6 field names. The values
# are one-argument functions that take a string and return an appropriately
# converted field value.
#
# The following fields are taken from the DIAMOND manual v0.9.22 2018-05-11.
# Fields whose name doesn't appear here will be left as strings.
DIAMOND_FIELD_CONVERTER = {
    "bitscore": float,
    "evalue": float,
    "frame": int,
    "gapopen": int,
    "gaps": int,
    "identicalCount": lambda nident: None if nident is None else int(nident),
    "length": int,
    "mismatch": int,
    "nident": int,
    "pident": float,
    "positive": int,
    "positiveCount": lambda pos: None if pos is None else int(pos),
    "ppos": float,
    "qcovhsp": float,
    "qend": int,
    "qframe": int,
    "qlen": int,
    "qstart": int,
    "score": float,
    "send": int,
    "slen": int,
    "sstart": int,
}


class DiamondTabularFormat:
    """
    Read/convert DIAMOND TAB-separated format (format #6).

    @param fieldNames: A C{list} or C{tuple} of C{str} DIAMOND field names.
        Run 'diamond -help' to see the full list. If C{None}, a default set of
        fields will be used, as compatible with convert-diamond-to-sam.py
    """

    def __init__(self, fieldNames=None):
        self._fieldNames = fieldNames or FIELDS.split()
        self._nFields = len(self._fieldNames)
        if not self._nFields:
            raise ValueError("fieldNames cannot be empty.")

        c = Counter(self._fieldNames)
        if c.most_common(1)[0][1] > 1:
            raise ValueError(
                "field names contains duplicated names: %s."
                % (", ".join(sorted(x[0] for x in c.most_common() if x[1] > 1)))
            )

        def identity(x):
            return x

        self._convertors = tuple(
            DIAMOND_FIELD_CONVERTER.get(field, identity) for field in self._fieldNames
        )

    def diamondFieldsToDict(self, line):
        """
        Convert a line of DIAMOND tabular (--outfmt 6) output to a dictionary

        @param line: A C{str} line read from DIAMOND output. This may contain
            a trailing whitespace (including newline and carriage return).
        @raise ValueError: If a line of C{filename} does not have the expected
            number of TAB-separated fields (i.e., len(self._fieldNames)).
        @return: A C{dict} with keys that are the DIAMOND field names, with
            values as converted by the functions set up in __init__.
        """
        values = line.rstrip().split("\t")

        if len(values) != self._nFields:
            raise ValueError(
                "DIAMOND output line had %d field values (expected %d). "
                "The offending input line was %r." % (len(values), self._nFields, line)
            )

        return dict(
            (fieldName, func(value))
            for fieldName, func, value in zip(
                self._fieldNames, self._convertors, values
            )
        )

    def diamondTabularFormatToDicts(self, filename):
        """
        Read DIAMOND tabular (--outfmt 6) output and convert lines to
        dictionaries.

        @param filename: Either a C{str} file name or an open file pointer.
        @raise ValueError: If a line of C{filename} does not have the expected
            number of TAB-separated fields (i.e., len(fieldNames)).
        @return: A generator that yields C{dict}s returned by
            C{diamondFieldsToDict}, above.
        """
        with as_handle(filename) as fp:
            for line in fp:
                yield self.diamondFieldsToDict(line)


class DiamondTabularFormatReader:
    """
    Provide a method that yields parsed tabular records from a file. Store and
    make accessible the global DIAMOND parameters.

    Make sure you run DIAMOND with the right output format. You must have run
    with --outfmt 6 and one of the following sets of arguments (in the given
    order):

        13 arg version:
        qtitle stitle bitscore evalue qframe qseq qstart qend sseq sstart send
            slen btop

        15 arg version:
        qtitle stitle bitscore evalue qframe qseq qstart qend sseq sstart send
            slen btop nident positive

        17 arg version:
        qtitle stitle bitscore evalue qframe qseq qstart qend sseq sstart send
            slen btop nident pident positive ppos

    @param filename: A C{str} filename or an open file pointer, containing
        DIAMOND tabular records.

    """

    def __init__(self, filename):
        self._filename = filename
        self.application = "DIAMOND"
        self.params = {
            "application": self.application,
            "reference": (
                "Buchfink et al., Fast and Sensitive Protein "
                "Alignment using DIAMOND, Nature Methods, 12, 59-60 "
                "(2015)"
            ),
            "task": "blastx",  # TODO: Add support for blastp, if needed.
            "version": "v0.8.23",
        }

    def records(self):
        """
        Parse the DIAMOND output and yield records. This will be used to read
        original DIAMOND output (either from stdin or from a file) to turn the
        DIAMOND results into Python dictionaries that will then be stored in
        our JSON format.

        @return: A generator that produces C{dict}s containing 'alignments' and
            'query' C{str} keys.
        """
        with as_handle(self._filename) as fp:
            previousQtitle = None
            subjectsSeen = {}
            record = {}
            for line in fp:
                line = line[:-1]
                fields = line.split("\t")
                nFields = len(fields)

                if nFields == 13:
                    (
                        qtitle,
                        stitle,
                        bitscore,
                        evalue,
                        qframe,
                        qseq,
                        qstart,
                        qend,
                        sseq,
                        sstart,
                        send,
                        slen,
                        btop,
                    ) = fields
                    nident = pident = positive = ppos = None
                elif nFields == 15:
                    (
                        qtitle,
                        stitle,
                        bitscore,
                        evalue,
                        qframe,
                        qseq,
                        qstart,
                        qend,
                        sseq,
                        sstart,
                        send,
                        slen,
                        btop,
                        nident,
                        positive,
                    ) = fields
                    pident = ppos = None
                elif nFields == 17:
                    (
                        qtitle,
                        stitle,
                        bitscore,
                        evalue,
                        qframe,
                        qseq,
                        qstart,
                        qend,
                        sseq,
                        sstart,
                        send,
                        slen,
                        btop,
                        nident,
                        pident,
                        positive,
                        ppos,
                    ) = fields
                else:
                    raise ValueError(
                        "Could not make sense of DIAMOND output. You must use "
                        "--outfmt 6 with 13, 15, or 17 arguments. See %s for "
                        "the expected names and order." % __file__
                    )

                hsp = {
                    "bits": float(bitscore),
                    "btop": btop,
                    "expect": float(evalue),
                    "frame": int(qframe),
                    "identicalCount": None if nident is None else int(nident),
                    "percentIdentical": (None if pident is None else float(pident)),
                    "positiveCount": (None if positive is None else int(positive)),
                    "percentPositive": (None if ppos is None else float(ppos)),
                    "query": qseq,
                    "query_start": int(qstart),
                    "query_end": int(qend),
                    "sbjct": sseq,
                    "sbjct_start": int(sstart),
                    "sbjct_end": int(send),
                }

                if previousQtitle == qtitle:
                    # We have already started accumulating alignments for this
                    # query.
                    if stitle not in subjectsSeen:
                        # We have not seen this subject before, so this is a
                        # new alignment.
                        subjectsSeen.add(stitle)
                        alignment = {
                            "hsps": [hsp],
                            "length": int(slen),
                            "title": stitle,
                        }
                        record["alignments"].append(alignment)
                    else:
                        # We have already seen this subject, so this is another
                        # HSP in an already existing alignment.
                        for alignment in record["alignments"]:
                            if alignment["title"] == stitle:
                                alignment["hsps"].append(hsp)
                                break
                else:
                    # All alignments for the previous query id (if any)
                    # have been seen.
                    if previousQtitle is not None:
                        yield record

                    # Start building up the new record.
                    record = {}
                    subjectsSeen = {stitle}
                    alignment = {
                        "hsps": [hsp],
                        "length": int(slen),
                        "title": stitle,
                    }
                    record["alignments"] = [alignment]
                    record["query"] = qtitle

                    previousQtitle = qtitle

            # Yield the last record, if any.
            if record:
                yield record

    def saveAsJSON(self, fp, writeBytes=False):
        """
        Write the records out as JSON. The first JSON object saved contains
        information about the DIAMOND algorithm.

        @param fp: A C{str} file pointer to write to.
        @param writeBytes: If C{True}, the JSON will be written out as bytes
            (not strings). This is required when we are writing to a BZ2 file.
        """
        if writeBytes:
            fp.write(dumps(self.params, sort_keys=True).encode("UTF-8"))
            fp.write(b"\n")
            for record in self.records():
                fp.write(dumps(record, sort_keys=True).encode("UTF-8"))
                fp.write(b"\n")
        else:
            fp.write(dumps(self.params, sort_keys=True))
            fp.write("\n")
            for record in self.records():
                fp.write(dumps(record, sort_keys=True))
                fp.write("\n")


class JSONRecordsReader:
    """
    Provide a method that yields JSON records from a file. Store, check, and
    make accessible the DIAMOND parameters.

    @param filename: A C{str} filename containing JSON DIAMOND records.
    @param scoreClass: A class to hold and compare scores (see scores.py).
        Default is C{HigherIsBetterScore}, for comparing bit scores. If you
        are using e-values, pass LowerIsBetterScore instead.
    """

    def __init__(self, filename, scoreClass=HigherIsBetterScore):
        self._filename = filename
        self._scoreClass = scoreClass
        if scoreClass is HigherIsBetterScore:
            self._hspClass = HSP
        else:
            self._hspClass = LSP

        self._open(filename)
        self.diamondTask = self.params["task"]

    def _open(self, filename):
        """
        Open the input file. Set self._fp to point to it. Read the first
        line of parameters.

        @param filename: A C{str} filename containing JSON DIAMOND records.
        @raise ValueError: if the first line of the file isn't valid JSON,
            if the input file is empty, or if the JSON does not contain an
            'application' key.
        """
        if filename.endswith(".bz2"):
            self._fp = bz2.open(filename, mode="rt", encoding="UTF-8")
        else:
            self._fp = open(filename)

        line = self._fp.readline()
        if not line:
            raise ValueError("JSON file %r was empty." % self._filename)

        try:
            self.params = loads(line[:-1])
        except ValueError as e:
            raise ValueError(
                "Could not convert first line of %r to JSON (%s). "
                "Line is %r." % (self._filename, e, line[:-1])
            )

    def _dictToAlignments(self, diamondDict, read):
        """
        Take a dict (made by DiamondTabularFormatReader.records)
        and convert it to a list of alignments.

        @param diamondDict: A C{dict}, from records().
        @param read: A C{Read} instance, containing the read that DIAMOND used
            to create this record.
        @return: A C{list} of L{dark.alignment.Alignment} instances.
        """
        alignments = []
        getScore = itemgetter("bits" if self._hspClass is HSP else "expect")

        for diamondAlignment in diamondDict["alignments"]:
            alignment = Alignment(diamondAlignment["length"], diamondAlignment["title"])
            alignments.append(alignment)
            for diamondHsp in diamondAlignment["hsps"]:
                score = getScore(diamondHsp)
                normalized = normalizeHSP(diamondHsp, len(read), self.diamondTask)
                hsp = self._hspClass(
                    score,
                    readStart=normalized["readStart"],
                    readEnd=normalized["readEnd"],
                    readStartInSubject=normalized["readStartInSubject"],
                    readEndInSubject=normalized["readEndInSubject"],
                    readFrame=diamondHsp["frame"],
                    subjectStart=normalized["subjectStart"],
                    subjectEnd=normalized["subjectEnd"],
                    readMatchedSequence=diamondHsp["query"],
                    subjectMatchedSequence=diamondHsp["sbjct"],
                    # Use diamondHsp.get on identicalCount, positiveCount,
                    # percentPositive, and percentIdentical because they
                    # were either added in version 2.0.3 or we didn't start
                    # using them until much later and so will not be
                    # present in any of our JSON output generated before
                    # that. Those values will be None when reading those
                    # JSON files, but that's much better than no longer
                    # being able to read all that earlier data.
                    identicalCount=diamondHsp.get("identicalCount"),
                    positiveCount=diamondHsp.get("positiveCount"),
                    percentIdentical=diamondHsp.get("percentIdentical"),
                    percentPositive=diamondHsp.get("percentPositive"),
                )

                alignment.addHsp(hsp)

        return alignments

    def readAlignments(self, reads):
        """
        Read lines of JSON from self._filename, convert them to read alignments
        and yield them.

        @param reads: An iterable of L{Read} instances, corresponding to the
            reads that were given to DIAMOND.
        @raise ValueError: If any of the lines in the file cannot be converted
            to JSON.
        @return: A generator that yields C{dark.alignments.ReadAlignments}
            instances.
        """
        if self._fp is None:
            self._open(self._filename)

        reads = iter(reads)

        try:
            for lineNumber, line in enumerate(self._fp, start=2):
                try:
                    record = loads(line[:-1])
                except ValueError as e:
                    raise ValueError(
                        "Could not convert line %d of %r to JSON (%s). "
                        "Line is %r." % (lineNumber, self._filename, e, line[:-1])
                    )
                else:
                    recordTitle = record["query"]
                    while True:
                        # Iterate through the input reads until we find the
                        # one that matches this DIAMOND record.
                        try:
                            read = next(reads)
                        except StopIteration:
                            raise ValueError(
                                "Read generator failed to yield a read "
                                "with id '%s' as found in record number %d "
                                "during parsing of DIAMOND output file %r."
                                % (recordTitle, lineNumber - 1, self._filename)
                            )
                        else:
                            # Look for an exact read id / subject title match.
                            # If that doesn't work, allow for the case where
                            # the JSON record has a truncated query (i.e.,
                            # read) id. This covers the situation where a tool
                            # we use (e.g., bwa mem) unconditionally does this
                            # truncation in the output it writes.
                            if (
                                read.id == recordTitle
                                or read.id.split()[0] == recordTitle
                            ):
                                alignments = self._dictToAlignments(record, read)
                                yield ReadAlignments(read, alignments)
                                break
                            else:
                                # This is an input read that had no DIAMOND
                                # matches. So it does not appear in the
                                # DIAMOND's output. Yield an empty
                                # ReadAlignments for it.
                                yield ReadAlignments(read, [])

        finally:
            self._fp.close()
            self._fp = None



#!/usr/bin/env python

import csv
import re
import sys
from typing import Union

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions

OVERLAPPING_INDICATOR = "+"

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            "Given FASTA on stdin and a sequence pattern, write information "
            "about which sequences and where the pattern matches to stdout."
        )
    )

    parser.add_argument(
        "--regex",
        required=True,
        help=(
            "The regular expression pattern to match. For matching, input "
            "sequences will be converted to uppercase unless --noUpper is used."
        ),
    )

    parser.add_argument(
        "--noUpper",
        dest="upper",
        action="store_false",
        help="Do not convert sequences to uppercase for the matching.",
    )

    parser.add_argument(
        "--overlapping", action="store_true", help="Include overlapping matches."
    )

    parser.add_argument(
        "--noId",
        dest="printId",
        action="store_false",
        help="Do not print sequence ids.",
    )

    parser.add_argument(
        "--noEnd",
        dest="printEnd",
        action="store_false",
        help="Do not print the end offset of matches.",
    )

    parser.add_argument(
        "--noSequence",
        dest="printSequence",
        action="store_false",
        help="Do not print sequence match regions.",
    )

    parser.add_argument(
        "--noOverlapping",
        dest="printOverlapping",
        action="store_false",
        help=(
            f"Do not print a {OVERLAPPING_INDICATOR!r} indicator to "
            f"identify matches that overlap the previously printed one."
        ),
    )

    parser.add_argument(
        "--zeroBased",
        action="store_true",
        help="Print half-open zero-based offsets, as in Python.",
    )

    parser.add_argument(
        "--count",
        "-c",
        action="store_true",
        help="Just print the count of the number of matches, if the count is non-zero.",
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    regex = re.compile(args.regex)
    upper = args.upper
    printEnd = args.printEnd
    printId = args.printId
    printSequence = args.printSequence
    printOverlapping = args.printOverlapping
    overlapping = args.overlapping
    startInc = int(not args.zeroBased)
    writerow = csv.writer(sys.stdout, delimiter="\t").writerow

    for read in reads:
        sequence = read.sequence.upper() if upper else read.sequence
        startPos = 0
        matches = []
        while True:
            match = regex.search(sequence, pos=startPos)
            if match:
                start, end = match.start(), match.end()
                startPos = start + 1 if overlapping else end
                matches.append(match)
            else:
                break

        if args.count:
            if matches:
                if printId:
                    writerow((read.id, len(matches)))
                else:
                    print(len(matches))
        else:
            lastEnd = -1
            for match in matches:
                start, end = match.start(), match.end()
                row: list[Union[int, str]] = [start + startInc]
                if printEnd:
                    row.append(end)
                if printId:
                    row.append(read.id)
                if printSequence:
                    row.append(match.group())
                if printOverlapping and start < lastEnd:
                    row.append(OVERLAPPING_INDICATOR)
                writerow(row)
                lastEnd = end



from io import BytesIO
from unittest import TestCase

import pytest

from dark.analyze_reads import getPrefixAndSuffix, trimReads


class TestGetPrefixAndSuffix(TestCase):
    """
    Tests for getPrefixAndSuffix()
    """

    def testZeroInput(self):
        result = getPrefixAndSuffix(BytesIO())
        self.assertEqual(result, (0, 0))

    def testOneInput(self):
        seq = b">hey\nagtcagtcagtc"
        result = getPrefixAndSuffix(BytesIO(seq))
        self.assertEqual(result, (12, 12))

    def testTwoDifferentReads(self):
        seq = b">hey\nagtcagtcagtc\n>you\ntcctg"
        result = getPrefixAndSuffix(BytesIO(seq))
        self.assertEqual(result, (0, 0))

    def testSamePrefixTwoReads(self):
        seq = b">hey\nagtcagtcagtc\n>you\nagttcctg"
        result = getPrefixAndSuffix(BytesIO(seq))
        self.assertEqual(result, (3, 0))

    def testSameSuffixTwoReads(self):
        seq = b">hey\nagtcagtcagtc\n>you\ntcctggtc"
        result = getPrefixAndSuffix(BytesIO(seq))
        self.assertEqual(result, (0, 3))

    def testSamePrefixAndSuffixTwoReads(self):
        seq = b">hey\nagtcagtcagtc\n>you\nagttcctggtc"
        result = getPrefixAndSuffix(BytesIO(seq))
        self.assertEqual(result, (3, 3))

    def testSamePrefixDifferentSuffixThreeReads(self):
        seq = b">hey\nagtcagtcagtc\n>you\nagttcctggtc\n>how\nagtcggtat"
        result = getPrefixAndSuffix(BytesIO(seq))
        self.assertEqual(result, (3, 0))

    def testSamePrefixSameSuffixThreeReads(self):
        seq = b">hey\nagtccttagatcg\n>you\nagtcgaatcg\n>how\nagtaacctcg"
        result = getPrefixAndSuffix(BytesIO(seq))
        self.assertEqual(result, (3, 3))

    def testDifferentPrefixSameSuffixThreeReads(self):
        seq = b">hey\nagtccttagatcg\n>you\ncgaatcg\n>how\natgacctcg"
        result = getPrefixAndSuffix(BytesIO(seq))
        self.assertEqual(result, (0, 3))


#
# Tests for trimReads()
#
@pytest.mark.parametrize(
    "sequence,prefix,suffix,expected",
    (
        (b"", 0, 0, ""),
        (b"", 10, 10, ""),
        (b"agtccgatcg", 0, 0, "agtccgatcg"),
        (b"agtccgatcg", 3, 0, "ccgatcg"),
        (b"agtccgatcg", 0, 3, "agtccga"),
        (b"agtccgatcg", 3, 3, "ccga"),
        (b"agtccgatcg", 8, 4, ""),
        (b"agtccgatcg", 4, 8, ""),
    ),
)
def testTrimReads(sequence, prefix, suffix, expected):
    result = list(trimReads(prefix, suffix, BytesIO(b">id\n" + sequence + b"\n")))[0]
    assert result.sequence == expected



#!/usr/bin/env python

import argparse
import sys

from dark import __version__
from dark.consensus import ConsensusError, consensusFromBAM
from dark.sam import SamError


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Make a consensus sequence.",
    )

    parser.add_argument(
        "--bamFilename",
        required=True,
        help="The BAM file from which the consensus should be called.",
    )

    parser.add_argument(
        "--version", action="store_true", help="Print the version number and exit"
    )

    parser.add_argument(
        "--bamId",
        metavar="ID",
        help=(
            "The BAM file reference name indicating which aligned "
            "reads to make a consensus from. If not given, will be inferred "
            "from the BAM file header."
        ),
    )

    parser.add_argument(
        "--referenceFasta", metavar="FILE", help="The reference FASTA file."
    )

    parser.add_argument(
        "--fastaId",
        metavar="ID",
        help=(
            "The id of the sequence in --referenceFasta to use as a "
            "reference. Only considered if --referenceFasta is used. If not "
            "given and --referenceFasta is, the reference id will be "
            "inferred from reference names in the BAM header, or will be "
            "taken as the id of the first sequence in --referenceFasta."
        ),
    )

    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        "--consensusId",
        metavar="ID",
        help=(
            "The id to use in the consensus sequence in the output FASTA. "
            "If not given, the reference sequence id will be used."
        ),
    )

    group.add_argument(
        "--idLambda",
        metavar='"lambda id: ..."',
        help=(
            "A one-argument function taking and returning a read id. "
            "This can be used to set the id of the consensus sequence based "
            "on the id of the reference sequence. The function will be "
            "called with the id of the BAM reference sequence. E.g., "
            "--idLambda \"lambda id: id.split('_')[0]\" or "
            "--idLambda \"lambda id: id[:10] + '-consensus'\"."
        ),
    )

    parser.add_argument(
        "--minCoverage",
        default=0,
        type=int,
        metavar="N",
        help=(
            "The minimum number of reads that must cover a site for a "
            "consensus base to be called. If zero reads cover a site, the "
            "--noCoverage value is used or if the number is greater than "
            "zero but less than minCoverage, the --lowCoverage value is "
            "used."
        ),
    )

    parser.add_argument(
        "--lowCoverage",
        default="n",
        metavar="N",
        help=(
            "What to do when some reads cover a site, but fewer than "
            'the --minCoverage value. Either set this to "reference" (to '
            "use the base from the reference sequence)or a single character "
            '(e.g., "N").'
        ),
    )

    parser.add_argument(
        "--noCoverage",
        default="N",
        help=(
            "What to do when no reads cover a site, Either set this to "
            '"reference" (to use the base from the reference sequence), or '
            'a single character (e.g., "N").'
        ),
    )

    parser.add_argument(
        "--threshold",
        type=float,
        default=0.6,
        metavar="N",
        help=(
            "The frequency threshold when calling the consensus. If the "
            "frequency of the most-common nucleotide at a site meets this "
            "threshold, that nucleotide will be called. Otherwise, an "
            "ambiguous nucleotide code will be produced, based on the "
            "smallest set of most-frequent nucleotides whose summed "
            "frequencies meet the threshold. If the frequency of the "
            "nucleotide that causes the threshold to be reached is the "
            "same as that of other nucleotides, all such nucleotides will "
            "be included in the ambiguous code."
        ),
    )

    parser.add_argument(
        "--ignoreQuality",
        action="store_true",
        help=(
            "Ignore FASTQ quality scores. All bases of all reads will be "
            "considered equally reliable."
        ),
    )

    parser.add_argument(
        "--includeSoftClipped",
        action="store_true",
        help=(
            "Include information from read bases that were marked as "
            "soft-clipped by the algorithm that made the BAM file. See "
            "https://samtools.github.io/hts-specs/SAMv1.pdf if this makes "
            "no sense to you."
        ),
    )

    parser.add_argument(
        "--compareWithPileupFile",
        metavar="FILE",
        help=(
            "Make a consensus using the pysam pileup function and compare "
            "it to the one made by usign pysam fetch. Write the summary to "
            'this file. If the file is "-", write to standard error.'
        ),
    )

    parser.add_argument(
        "--deletionSymbol",
        default="-",
        metavar="CHAR",
        help=(
            "When a deletion is detected, put this symbol into the "
            "consensus sequence. Use the empty string to have deletions "
            "omitted."
        ),
    )

    parser.add_argument(
        "--strategy",
        default="fetch",
        choices=("fetch",),
        help="The consensus-making strategy to use.",
    )

    parser.add_argument(
        "--deletionThreshold",
        default=0.5,
        type=float,
        metavar="N",
        help=(
            "If some reads have a deletion at a site and some do not, call "
            "the site as a deletion if the fraction of reads with the "
            "deletion is at least this value."
        ),
    )

    parser.add_argument(
        "--insertionCountThreshold",
        default=5,
        type=int,
        metavar="N",
        help=(
            "The number of reads that must have an insertion at an offset "
            "in order for the insertion to be called in the consensus."
        ),
    )

    parser.add_argument(
        "--progress",
        action="store_true",
        help=("Show a progress bar (unless standard error has been redirected)."),
    )

    parser.add_argument(
        "--quiet",
        action="store_true",
        help=(
            "Suppress diagnostic output. Note that this will silence "
            "warnings about differing reference names."
        ),
    )

    args = parser.parse_args()

    if args.version:
        print(__version__)
        sys.exit(0)

    try:
        consensus = consensusFromBAM(
            args.bamFilename,
            bamId=args.bamId,
            referenceFasta=args.referenceFasta,
            fastaId=args.fastaId,
            consensusId=args.consensusId,
            idLambda=args.idLambda,
            threshold=args.threshold,
            minCoverage=args.minCoverage,
            lowCoverage=args.lowCoverage,
            noCoverage=args.noCoverage,
            deletionSymbol=args.deletionSymbol,
            deletionThreshold=args.deletionThreshold,
            insertionCountThreshold=args.insertionCountThreshold,
            ignoreQuality=args.ignoreQuality,
            includeSoftClipped=args.includeSoftClipped,
            strategy=args.strategy,
            compareWithPileupFile=args.compareWithPileupFile,
            progress=args.progress,
            quiet=args.quiet,
        )
    except (ConsensusError, SamError) as e:
        print(f"{e} Exiting.", file=sys.stderr)
        sys.exit(1)
    else:
        print(consensus.toString("fasta"), end="")


if __name__ == "__main__":
    main()



from unittest import TestCase

from dark.btop import btop2cigar, countGaps, parseBtop


def btop2cigarStr(btopString, concise=False, aa=False):
    """
    Call btopString and turn its generator result into a string.

    See btopString for the explanation of the args.
    """
    return "".join(btop2cigar(btopString, concise, aa))


class TestParseBtop(TestCase):
    """
    Tests for the parseBtop function.
    """

    def testOneLetter(self):
        """
        An argument with just one letter must result in a ValueError.
        """
        error = (
            "^BTOP string 'F' has a trailing query letter 'F' with no "
            "corresponding subject letter$"
        )
        self.assertRaisesRegex(ValueError, error, list, parseBtop("F"))

    def testConsecutiveGaps(self):
        """
        An argument that has two consecutive gaps characters must result in a
        ValueError.
        """
        error = "^BTOP string '36--' has two consecutive gaps at offset 2$"
        self.assertRaisesRegex(ValueError, error, list, parseBtop("36--"))

    def testConsecutiveIdentical(self):
        """
        An argument that has two consecutive identical (non-gap) characters
        must result in a ValueError.
        """
        error = (
            "^BTOP string '36AA' has two consecutive identical 'A' letters at offset 2$"
        )
        self.assertRaisesRegex(ValueError, error, list, parseBtop("36AA"))

    def testEmpty(self):
        """
        An empty argument must result in an empty list.
        """
        self.assertEqual([], list(parseBtop("")))

    def testOneNumberWithTrailingOneLetter(self):
        """
        An argument that is a number with a single letter must result in a
        ValueError.
        """
        error = (
            "^BTOP string '36F' has a trailing query letter 'F' with no "
            "corresponding subject letter$"
        )
        self.assertRaisesRegex(ValueError, error, list, parseBtop("36F"))

    def testThreeLetters(self):
        """
        An argument that has three letters must result in a ValueError.
        """
        error = (
            "^BTOP string 'ABC' has a trailing query letter 'C' with no "
            "corresponding subject letter$"
        )
        self.assertRaisesRegex(ValueError, error, list, parseBtop("ABC"))

    def testOneLetterThenANumber(self):
        """
        An argument that is a single letter followed by a number must result
        in a ValueError.
        """
        error = (
            "^BTOP string 'F36' has a query letter 'F' at offset 0 with "
            "no corresponding subject letter$"
        )
        self.assertRaisesRegex(ValueError, error, list, parseBtop("F36"))

    def testTwoNumbersWithOneLetterBetween(self):
        """
        An argument that is a number, a single letter, and another number must
        result in a ValueError.
        """
        error = (
            "^BTOP string '36F77' has a query letter 'F' at offset 2 "
            "with no corresponding subject letter$"
        )
        self.assertRaisesRegex(ValueError, error, list, parseBtop("36F77"))

    def testOneNumber(self):
        """
        An argument that is just one number must give the expected result.
        """
        self.assertEqual([54], list(parseBtop("54")))

    def testOneNumberThatIsZero(self):
        """
        An argument that is just the number zero must give the expected result.
        """
        self.assertEqual([0], list(parseBtop("0")))

    def testOneNumberWithLeadingZeroes(self):
        """
        An argument that is just one number with leading zeroes must give the
        expected result.
        """
        self.assertEqual([54], list(parseBtop("0054")))

    def testOneQuerySubjectPair(self):
        """
        An argument that is a single query/subject letter pair must give the
        expected result.
        """
        self.assertEqual([("A", "G")], list(parseBtop("AG")))

    def testTwoQuerySubjectPairs(self):
        """
        An argument that has two query/subject letter pairs must give the
        expected result.
        """
        self.assertEqual([("A", "G"), ("C", "T")], list(parseBtop("AGCT")))

    def testOneQuerySubjectPairAndANumber(self):
        """
        An argument that is a single query/subject letter pair followed by a
        number must give the expected result.
        """
        self.assertEqual([("A", "G"), 33], list(parseBtop("AG33")))


class TestCountGaps(TestCase):
    """
    Tests for the countGaps function.
    """

    def testEmpty(self):
        """
        An argument with an empty string must produce the expected result.
        """
        self.assertEqual((0, 0), countGaps(""))

    def testNumberOnly(self):
        """
        An argument with just a number must produce the expected result.
        """
        self.assertEqual((0, 0), countGaps("88"))

    def testLettersButNoGaps(self):
        """
        An argument with just letters must produce the expected result.
        """
        self.assertEqual((0, 0), countGaps("FGAC"))

    def testOneQueryGap(self):
        """
        An argument with just a query gap must produce the expected result.
        """
        self.assertEqual((1, 0), countGaps("-G"))

    def testOneSubjectGap(self):
        """
        An argument with just a subject gap must produce the expected result.
        """
        self.assertEqual((0, 1), countGaps("G-"))

    def testOneQueryAndOneSubjectGap(self):
        """
        An argument with a query and a subject gap must produce the expected
        result.
        """
        self.assertEqual((1, 1), countGaps("G--G"))

    def testMultipleQueryAndSubjectGaps(self):
        """
        An argument with multiple query and a subject gaps must produce the
        expected result.
        """
        self.assertEqual((3, 2), countGaps("-GG-34-T-T39F-"))


class TestBtop2CigarPrecise(TestCase):
    """
    Tests for the btop2cigar function when concise is False.
    """

    def testEmpty(self):
        """
        An empty BTOP string must result in an empty CIGAR string.
        """
        self.assertEqual("", btop2cigarStr("", concise=False))

    def testMixedMatch(self):
        """
        If a BTOP string specifies that all characters match (in the imprecise
        CIGAR sense where M could be identical characters or not), the CIGAR
        string must be all Ms.
        """
        self.assertEqual("7M", btop2cigarStr("2GC3AT", concise=False))

    def testRefenceInsertion(self):
        """
        If a BTOP string specifies that the query has character but the
        subject (reference) does not, the CIGAR string must indicate an
        insertion to the reference.
        """
        self.assertEqual("1I", btop2cigarStr("A-", concise=False))

    def testQueryInsertion(self):
        """
        If a BTOP string specifies that the subject (reference) has character
        but the query does not, the CIGAR string must indicate an deletion in
        the reference.
        """
        self.assertEqual("1D", btop2cigarStr("-A", concise=False))

    def testAll(self):
        """
        If a BTOP string specifies all possible variations, the CIGAR
        string must be correct.
        """
        self.assertEqual("7M2I4M2D5M", btop2cigarStr("2GC3ATC-G-4-T-A5", concise=False))

    def testAllList(self):
        """
        If a BTOP string specifies all possible variations, the CIGAR
        string components must be correct.
        """
        self.assertEqual(
            ["7M", "2I", "4M", "2D", "5M"],
            list(btop2cigar("2GC3ATC-G-4-T-A5", concise=False)),
        )

    def testAllAA(self):
        """
        If a BTOP string specifies all possible variations, and we indicate
        that the BTOP string refers to amino acids, the CIGAR string must be
        correct (i.e., all counts must be tripled).
        """
        self.assertEqual(
            "21M6I12M6D15M", btop2cigarStr("2GC3ATC-G-4-T-A5", concise=False, aa=True)
        )

    def testAllAAList(self):
        """
        If a BTOP string specifies all possible variations, and we indicate
        that the BTOP string refers to amino acids, the CIGAR string components
        must be correct (i.e., all counts must be tripled).
        """
        self.assertEqual(
            ["21M", "6I", "12M", "6D", "15M"],
            list(btop2cigar("2GC3ATC-G-4-T-A5", concise=False, aa=True)),
        )


class TestBtop2CigarConcise(TestCase):
    """
    Tests for the btop2cigar function when concise is True.
    """

    def testEmpty(self):
        """
        An empty BTOP string must result in an empty CIGAR string.
        """
        self.assertEqual("", btop2cigarStr("", concise=True))

    def testMixedMatch(self):
        """
        If a BTOP string specifies that some characters match and some do
        not, the CIGAR string must be specific about the matches / mismatches.
        """
        self.assertEqual("2=1X3=1X", btop2cigarStr("2GC3AT", concise=True))

    def testRefenceInsertion(self):
        """
        If a BTOP string specifies that the query has character but the
        subject (reference) does not, the CIGAR string must indicate an
        insertion to the reference.
        """
        self.assertEqual("1I", btop2cigarStr("A-", concise=True))

    def testQueryInsertion(self):
        """
        If a BTOP string specifies that the subject (reference) has character
        but the query does not, the CIGAR string must indicate an deletion in
        the reference.
        """
        self.assertEqual("1D", btop2cigarStr("-A", concise=True))

    def testAll(self):
        """
        If a BTOP string specifies all possible variations, the CIGAR
        string must be correct.
        """
        self.assertEqual(
            "2=1X3=1X2I4=2D5=", btop2cigarStr("2GC3ATC-G-4-T-A5", concise=True)
        )

    def testWithAATrue(self):
        """
        If concise and aa are both set to True, a ValueError must be raised.
        """
        error = "^aa and concise cannot both be True$"
        self.assertRaisesRegex(
            ValueError, error, btop2cigarStr, "", concise=True, aa=True
        )



import sys
from collections import defaultdict

from pysam import CDEL, CINS, CSOFT_CLIP

from dark.cigar import cigarTuplesToOperations, insertionOffset, softClippedOffset
from dark.dna import Bases
from dark.progress import maybeProgressBar
from dark.reads import DNARead
from dark.sam import CONSUMES_REFERENCE, UnspecifiedReference, getReferenceInfo, samfile
from dark.utils import pct

DEBUG = False


def debug(*msg):
    print(*msg, file=sys.stderr)


class ConsensusError(Exception):
    "A consensus-making error."


class Insertion:
    """
    Manage a collection of bases and quality scores for an insertion that
    starts at given genome site.

    @param insertionOffset: The C{int} offset of the insertion.
    """

    def __init__(self, insertionOffset):
        self.insertionOffset = insertionOffset
        self.anchorOffsets = []
        self.bases = []

    def __str__(self):
        result = []

        for anchorOffset, bases in zip(self.anchorOffsets, self.bases):
            result.append(f"  {anchorOffset}: {bases}")

        return f"<Insertion at {self.insertionOffset}:\n" + "\n".join(result) + ">"

    __repr__ = __str__

    def start(self, anchorOffset):
        """
        Start a new insertion.

        @param anchorOffset: The C{int} offset where the insertion should begin
            (in the reference). Or C{None} if the insertion should be placed
            immediately before {self.insertionOffset}.
        """
        self.anchorOffsets.append(anchorOffset)
        self.bases.append([])

    def readCount(self):
        """
        How many reads have an insertion at this offset?

        @return: An C{int} read count.
        """
        return len(self.bases)

    def append(self, base, quality):
        """
        Append a base and quality score to the current list.

        @param base: A C{str} nucleotide base.
        @param quality: An C{int} nucleotide quality score.
        @raise IndexError: If C{start} has not already been called.
        """
        self.bases[-1].append((base, quality))

    def resolve(self):
        """
        Resolve the insertion at this offset.

        @return: A C{list} of Bases() instances corresponding to the insertion.
        """
        # The length of the insertion will be at least the length of the
        # longest sequence of (base, quality) pairs we've been given for
        # this offset.
        nInsertions = max(len(bases) for bases in self.bases)

        # And we have to potentially add so that if some reads need to be
        # anchored at a certain starting offset.
        maxExtra = 0
        for anchorOffset, bases in zip(self.anchorOffsets, self.bases):
            if anchorOffset is not None:
                extra = (
                    self.insertionOffset + nInsertions - anchorOffset - len(bases) - 1
                )
                if DEBUG:
                    debug(
                        f"{extra} = {self.insertionOffset} + "
                        f"{nInsertions} - {anchorOffset} - {len(bases)} - 1"
                    )
                if extra > maxExtra:
                    maxExtra = extra

        nInsertions += maxExtra

        # Make an array of Bases instances and add the (base, quality) pairs
        # to each one.
        insertion = []
        for _ in range(nInsertions):
            insertion.append(Bases())

        for anchorOffset, bases in zip(self.anchorOffsets, self.bases):
            startOffset = nInsertions - len(bases) if anchorOffset is None else 0
            for offset, (base, quality) in enumerate(bases):
                insertion[startOffset + offset].append(base, quality)

        if DEBUG:
            debug("insertion")
            for i in insertion:
                debug("\t", i)

        return insertion


def basesToConsensus(
    offsetBases,
    otherBases,
    originalOffsets,
    reference,
    referenceLength,
    threshold,
    minCoverage,
    lowCoverage,
    noCoverage,
    deletionSymbol,
    progress,
):
    """
    Convert a C{dict} of C{Bases} into a consensus.

    @param offsetBases: A C{defaultdict} mapping C{int} offsets to
        C{Bases} instances.
    @param otherBases: A C{dict} mapping C{int} offsets to C{str} symbols
        for the consensus for offsets that we have no reads for.
    @param originalOffsets: A C{dict} mapping C{int} adjusted offsets to their
        C{int} original offsets.
    @param reference: A C{Read} instance giving the reference sequence, or
        C{None} (in which case C{lowCoverage} and C{noCoverage} may not be
        'reference'.
    @param referenceLength: The C{int} length of the reference (note that we
        may not have the reference sequence but we can still be passed its
        length, as found in the BAM file header).
    @param threshold: A C{float} threshold, as for C{consensusFromBAM}.
    @param minCoverage: An C{int} minimum number of reads that must cover a
        site for a consensus base to be called. If fewer reads cover a
        site, the C{lowCoverage} value is used.
    @param lowCoverage: A C{str} indicating what base to use when
        0 < N < minCoverage reads cover a site.
    @param noCoverage: A C{str} indicating what base to use when
        no reads cover the site.
    @parm deletionSymbol: The C{str} to insert in the consensus when a deleted
        site is detected.
    @param progress: If C{True}, display a progress bar on standard error.
    @return: a C{str} consensus sequence.
    """
    if DEBUG:
        debug("In basesToConsensus: offsetBases:")
        for offset in offsetBases:
            debug(f"  {offset}: {offsetBases[offset]}")
        debug("In basesToConsensus: otherBases:")
        for offset in otherBases:
            debug(f"  {offset}: {otherBases[offset]}")

    result = []
    allOffsets = set(offsetBases) | set(otherBases) | {0}
    minOffset = min(allOffsets)
    maxOffset = max(allOffsets)

    noCoverageStr = (
        reference.sequence
        if noCoverage == "reference"
        else noCoverage * referenceLength
    )
    lowCoverageStr = (
        reference.sequence
        if lowCoverage == "reference"
        else lowCoverage * referenceLength
    )

    with maybeProgressBar(progress, maxOffset - minOffset + 1, "Consensus: ") as bar:
        for barCount, offset in enumerate(range(minOffset, maxOffset + 1), start=1):
            try:
                originalOffset = originalOffsets[offset]
                lowCoverageBase = lowCoverageStr[originalOffset]
                noCoverageBase = noCoverageStr[originalOffset]
            except (IndexError, KeyError):
                # TODO: Check this out. Add a variable to set this or... ?
                lowCoverageBase = noCoverageBase = "?"

            if offset in otherBases and (
                offset not in offsetBases or not offsetBases[offset]
            ):
                result.append(
                    deletionSymbol if otherBases[offset] is None else otherBases[offset]
                )
            else:
                assert offset in offsetBases, (
                    f"Offset {offset} not found in offsetBases or otherBases."
                )
                bases = offsetBases[offset]
                result.append(
                    bases.consensus(
                        threshold, minCoverage, lowCoverageBase, noCoverageBase
                    )
                )

            bar.update(barCount)

    return "".join(result)


def getConsensusId(bamId, idLambda):
    """
    Make an id for the consensus sequence.

    @param bamId: A C{str} reference sequence id from the BAM file.
    @param idLambda: A one-argument function taking and returning a sequence
        id. This can be used to set the id of the consensus sequence based
        on the id of the reference sequence. The function will be called with
        the id of the BAM reference sequence.
    @return: A C{str} sequence id for the consensus.
    """
    if idLambda is None:
        consensusId = f"{bamId}-consensus"
    else:
        idLambda = eval(idLambda)
        consensusId = idLambda(bamId)

    return consensusId


def consensusFromBAM(
    bamFilename,
    bamId=None,
    referenceFasta=None,
    fastaId=None,
    consensusId=None,
    idLambda=None,
    threshold=0.8,
    minCoverage=1,
    lowCoverage="reference",
    noCoverage="reference",
    deletionSymbol="-",
    deletionThreshold=0.5,
    ignoreQuality=False,
    insertionCountThreshold=5,
    strategy="fetch",
    includeSoftClipped=False,
    compareWithPileupFile=None,
    progress=False,
    quiet=False,
):
    """
    Build a consensus sequence from a BAM file.

    @param bamFilename: the BAM file.
    @param bamId: A C{str} BAM file reference name indicating which aligned
        reads to make a consensus from. If not given, will be inferred
        from the BAM file header.
    @param referenceFasta: A C{str} file name containing the sequence that was
        aligned to in making the BAM file.
    @param fastaId: A C{str} reference name indicating which sequence in
        C{referenceFasta} to use as a reference. Only considered if
        C{referenceFasta} is given. If not given and C{referenceFasta} is,
        the reference id will be inferred from reference names in the BAM
        header, or will be taken as the id of the first sequence in
        C{referenceFasta}.
    @param consensusId: The C{str} id to use in the consensus sequence. If not
        given, the BAM reference id with '-consensus' appended will be used.
    @param idLambda: A one-argument function taking and returning a sequence
        id. This can be used to set the id of the consensus sequence based
        on the id of the reference sequence. The function will be called with
        the id of the BAM reference sequence.
    @param threshold: A C{float} threshold. This fraction, at least, of the
        most-common nucleotides at a site are used to determine the consensus
        nucleotide (or ambiguous symbol if more than one nucleotide is
        required to achieve this threshold). If there is a tie in nucleotide
        counts at a site that causes the threshold to be met, all nucleotides
        of equeal frequncy will be included in the ambiguous symbol for that
        site. This is perhaps better explained with an example. See
        https://assets.geneious.com/manual/2020.1/static/GeneiousManualse43.html
        and the corresponding testGeneiousExamplesTie test in test/test_dna.py
    @param minCoverage: An C{int} minimum number of reads that must cover a
        site for a consensus base to be called. If zero reads cover a site, the
        C{noCoverage} value is used or if the number is greater than zero but
        less than C{minCoverage}, the C{lowCoverage} value is used.
    @param lowCoverage: A C{str} indicating what to do when some reads cover a
        site, but fewer than C{minCoverage}. Either 'reference' or a single
        character (e.g., 'N').
    @param noCoverage: A C{str} indicating what to do when no reads cover a
        reference base. Either 'reference' or a single character (e.g., 'N').
    @parm deletionSymbol: The C{str} to insert in the consensus when a deleted
        site is detected.
    @param deletionThreshold: If some reads have a deletion at a site and some
        do not, call the site as a deletion if C{float} the fraction of reads
        with the deletion is at least this value.
    @param insertionCountThreshold: The C{int} number of reads that must have
        an insertion at an offset in order for the insertion to be called in
        the consensus.
    @param ignoreQuality: If C{True}, ignore quality scores.
    @param strategy: A C{str} consensus-making strategy.
    @param includeSoftClipped: Include information from read bases that were
        marked as soft-clipped by the algorithm that made the BAM file.
    @param compareWithPileupFile: If C{True}, compare the base counts from the
        pysam fetch method with those of the pileup methods. This pays no
        attention to insertions. A summary of the result is written to this
        file.
    @param progress: If C{True}, display a progress bar on standard error.
    @param quiet: If C{True}, suppress diagnostic output. Note that this will
        silence warnings about differing reference names.
    @raise UnspecifiedReference: If no id is provided to indicate which BAM
        file reference to call a consensus for.
    @raise UnknownReference: If a requested reference id is unknown.
    @raise UnequalReferenceLengthError: If the passed reference does not have a
        length identical to the length mentioned in the BAM file.
    @raise ReferenceNameMismatchError: If the name of the FASTA reference
        sequence and the BAM reference do not agree (this is not raised if both
        ids are given explicitly).
    @return: A C{Read} instance with the consensus sequence.
    """
    if referenceFasta is None:
        if lowCoverage == "reference":
            raise UnspecifiedReference(
                'lowCoverage is "reference" but no reference FASTA file was given.'
            )

        if noCoverage == "reference":
            raise UnspecifiedReference(
                'noCoverage is "reference" but no reference FASTA file was given.'
            )

    with samfile(bamFilename) as bam:
        bamId, reference, referenceLength = getReferenceInfo(
            bam, bamFilename, bamId, referenceFasta, fastaId, quiet
        )

        if consensusId is None:
            consensusId = getConsensusId(bamId, idLambda)

        correspondences, deletions, insertions = getPairs(
            bam, bamId, referenceLength, ignoreQuality, includeSoftClipped, progress
        )

        conflicts = set(deletions) & set(insertions)
        if conflicts and DEBUG:
            debug("CONFLICTING OFFSETS:", ", ".join(sorted(map(str, conflicts))))

        if strategy == "fetch":
            (
                correspondences,
                consensusBases,
                otherBases,
                originalOffsets,
            ) = fetchConsensus(
                bam,
                correspondences,
                deletions,
                insertions,
                reference,
                referenceLength,
                noCoverage,
                deletionThreshold,
                ignoreQuality,
                insertionCountThreshold,
                includeSoftClipped,
                progress,
            )
        else:
            raise ConsensusError(f"Unknown consensus strategy {strategy!r}.")

        if compareWithPileupFile:
            with open(compareWithPileupFile, "w") as fp:
                compareCorrespondences(
                    fp,
                    correspondences,
                    pileupCorrespondences(
                        bam, bamId, referenceLength, includeSoftClipped, progress
                    ),
                    threshold,
                    minCoverage,
                )

    consensus = basesToConsensus(
        consensusBases,
        otherBases,
        originalOffsets,
        reference,
        referenceLength,
        threshold,
        minCoverage,
        lowCoverage,
        noCoverage,
        deletionSymbol,
        progress,
    )

    return DNARead(consensusId, consensus)


def getPairs(bam, bamId, referenceLength, ignoreQuality, includeSoftClipped, progress):
    """Compute a majority consensus using pysam's fetch function.

    @param bam: An open BAM file.
    @param bamId: A C{str} reference name indicating which reference to
        find reads against in the BAM file. If C{None} and there is only one
        reference in the BAM file, that one will be used, else a ConsensusError
        is raised due to not knowing which reference to use.
    @param reference: A C{Read} instance giving the reference sequence, or
        C{None}.
    @param referenceLength: The C{int} length of the reference (note that we
        may not have the reference sequence but we can still get its length
        from the BAM file header).
    @param ignoreQuality: If C{True}, ignore quality scores.
    @param includeSoftClipped: If C{True} include information from read bases
        marked as soft-clipped by the algorithm that made the BAM file.
    @param progress: If C{True}, display a progress bar on standard error.
    @return: A 3-tuple of correspondences, deletions, insertions, with types
        as in the lines just below.
    """
    correspondences = defaultdict(lambda: Bases())
    deletions = defaultdict(int)
    insertions = {}
    nReads = bam.count(contig=bamId)

    with maybeProgressBar(progress, nReads, "Reads    : ") as bar:
        for readCount, read in enumerate(bam.fetch(contig=bamId), start=1):
            assert not read.is_unmapped

            addPairsInfo(
                read.get_aligned_pairs(),
                list(cigarTuplesToOperations(read.cigartuples)),
                read.query_sequence,
                (
                    [1] * len(read.query_sequence)
                    if ignoreQuality
                    else read.query_qualities
                ),
                referenceLength,
                includeSoftClipped,
                correspondences,
                deletions,
                insertions,
            )

            if DEBUG:
                debug(f"read id     : {read.query_name}")
                debug("query       :", read.query_sequence)
                debug("query len   :", len(read.query_sequence))
                debug(f"cigar       : {read.cigarstring}")
                debug(f"match offset: {read.reference_start}")
                debug(f"Pairs       : {read.get_aligned_pairs()}")
                debug(f"Deletions   : {deletions}")

                debug("correspondences:")
                for offset, item in sorted(correspondences.items()):
                    debug(f"  {offset}: {item}")

                debug("insertions:")
                for offset in sorted(insertions):
                    debug(f"  {insertions[offset]}")

            bar.update(readCount)

    return correspondences, deletions, insertions


def addPairsInfo(
    pairs,
    cigarOperations,
    query,
    qualities,
    referenceLength,
    includeSoftClipped,
    correspondences,
    deletions,
    insertions,
):
    """
    Add information about matched pairs of nucleotides.

    @param pairs: A C{list} of 2-C{tuple}s of query offset, reference offset.
        Either (but not both) member of each tuple might be C{None} to indicate
        an indel mismatch.
    @param cigarOperations: A C{list} of CIGAR operations corresponding to the
        information in C{pairs}.
    @param query: A C{str} query DNA sequence.
    @param qualities: A C{list} of quality scores.
    @param includeSoftClipped: Include information from read bases that were
        marked as soft-clipped by the algorithm that made the BAM file.
    @param correspondences: A C{defaultdict(list)}, to hold (base, quality)
        scores for when a query offset corresponds to a reference offset.
    @param deletions: A C{set} of C{int} reference offsets that are deleted in
        the query.
    @param insertions: A C{defaultdict(list)}, to hold (base, quality)
        scores for when a query contains an insertion to the reference.
    """
    assert len(pairs) == len(cigarOperations)
    assert not any(pair == (None, None) for pair in pairs)

    # If the first reference offset is None, the first operation must be a
    # soft clip (not an insertion).
    if pairs[0][1] is None:
        assert cigarOperations[0] == CSOFT_CLIP, (
            f"First CIGAR operation is {cigarOperations[0]}, not a soft clip"
        )

    inInsertion = False

    for count, ((queryOffset, referenceOffset), cigarOperation) in enumerate(
        zip(pairs, cigarOperations)
    ):
        if queryOffset is None:
            # The query is missing something that is in the reference. So this
            # is a deletion from the reference.
            assert cigarOperation == CDEL
            assert referenceOffset is not None
            deletions[referenceOffset] += 1
            inInsertion = False

        elif referenceOffset is None:
            base = query[queryOffset]
            quality = qualities[queryOffset]

            if cigarOperation == CINS:
                # The query has an insertion (relative to the reference).

                # A CIGAR string shouldn't start with an insertion, IMO.
                # Rather, in such a case, it must start with unmatched
                # (soft-clipped) bases.
                # assert lastReferenceOffset is not None

                lookedBack, iOffset = insertionOffset(count, pairs, cigarOperations)
                if not inInsertion:
                    inInsertion = True
                    if iOffset not in insertions:
                        insertions[iOffset] = Insertion(iOffset)
                    insertions[iOffset].start(iOffset if lookedBack else None)

                insertions[iOffset].append(base, quality)
            else:
                assert cigarOperation == CSOFT_CLIP
                inInsertion = False
                if includeSoftClipped:
                    correspondences[
                        softClippedOffset(count, pairs, cigarOperations)
                    ].append(base, quality)
        else:
            # Query and reference offsets are both non-None.
            assert cigarOperation in CONSUMES_REFERENCE
            inInsertion = False
            base = query[queryOffset]
            quality = qualities[queryOffset]
            correspondences[referenceOffset].append(base, quality)


def fetchConsensus(
    bam,
    correspondences,
    deletions,
    insertions,
    reference,
    referenceLength,
    noCoverage,
    deletionThreshold,
    ignoreQuality,
    insertionCountThreshold,
    includeSoftClipped,
    progress,
):
    """Compute a majority consensus using pysam's fetch function.

    @param bam: An open BAM file.
    @param bamId: A C{str} reference name indicating which reference to
        find reads against in the BAM file. If C{None} and there is only one
        reference in the BAM file, that one will be used, else a ConsensusError
        is raised due to not knowing which reference to use.
    @param reference: A C{Read} instance giving the reference sequence, or
        C{None}.
    @param referenceLength: The C{int} length of the reference (note that we
        may not have the reference sequence but we can still get its length
        from the BAM file header).
    @param noCoverage: A C{str} indicating what to do when no reads cover a
        reference base. Either 'reference' or a single character (e.g., 'N').
    @param deletionThreshold: If some reads have a deletion at a site and some
        do not, call the site as a deletion if C{float} the fraction of reads
        with the deletion is at least this value.
    @param insertionCountThreshold: The C{int} number of reads that must have
        an insertion at an offset in order for the insertion to be called in
        the consensus.
    @param ignoreQuality: If C{True}, ignore quality scores.
        to.
    @param includeSoftClipped: Include information from read bases that were
        marked as soft-clipped by the algorithm that made the BAM file.
    @param progress: If C{True}, display a progress bar on standard error.
    @return: A 4-C{tuple} of correspondences, consensusBases, otherBases,
        and originalOffsets (as below).

    """
    # Do deletions.
    if DEBUG and deletions:
        debug(f"There are {len(deletions)} deletions:")

    actualDeletions = set()
    for offset, deletionCount in sorted(deletions.items()):
        count = correspondences[offset].count
        if count == 0 or deletionCount / count >= deletionThreshold:
            actualDeletions.add(offset)
            if DEBUG:
                debug(
                    f"  Offset {offset:5d}: deleted in {deletionCount:4d} "
                    f"present in {count:4d}"
                )

    if DEBUG:
        debug(f"Actual deletions: {actualDeletions}")
        debug("correspondences after deletions:")
        for offset, item in sorted(correspondences.items()):
            debug(f"  {offset}: {item}")

    # Do insertions.

    if DEBUG:
        debug(f"There are {len(insertions)} insertions.")

    allOffsets = set(correspondences) | {0, referenceLength - 1}
    minOffset = min(allOffsets)
    maxOffset = max(allOffsets)
    consensusBases = defaultdict(lambda: Bases())
    otherBases = {}
    originalOffsets = {}

    if DEBUG:
        debug(f"Offset limits: {minOffset} - {maxOffset}")
        debug(f"Insertion count {len(insertions)}")
        debug("final insertions:")
        for offset in sorted(insertions):
            wanted = insertions[offset].readCount() >= insertionCountThreshold
            debug(f"  {'' if wanted else 'un'}wanted {insertions[offset]}")

    with maybeProgressBar(progress, maxOffset - minOffset + 1, "Collect  : ") as bar:
        insertCount = 0
        for barCount, offset in enumerate(range(minOffset, maxOffset + 1), start=1):
            if offset in insertions:
                insertion = insertions[offset]
                assert offset == insertion.insertionOffset
                if insertion.readCount() >= insertionCountThreshold:
                    insertionBases = insertion.resolve()
                    insertionLength = len(insertionBases)
                    if DEBUG:
                        debug("INSERTION:", insertion)
                        debug(f"  Resolved {insertionLength=}")
                        debug(f"{insertionBases=}")

                    for bases in insertionBases:
                        adjustedOffset = offset + insertCount
                        consensusBases[adjustedOffset] += bases
                        insertCount += 1
                        originalOffsets[adjustedOffset] = offset

            adjustedOffset = offset + insertCount
            originalOffsets[adjustedOffset] = offset

            if offset in actualDeletions:
                otherBases[adjustedOffset] = None

            if offset in correspondences:
                if DEBUG:
                    debug(
                        f"Offset {offset} Adding correspondences bases "
                        f"{correspondences[offset]} to "
                        f"{consensusBases[adjustedOffset]} at offset "
                        f"{adjustedOffset}"
                    )
                consensusBases[adjustedOffset] += correspondences[offset]

            if adjustedOffset not in consensusBases:
                # We don't have any information (from the reads) for this
                # offset. If it's present in otherBases, it must be because
                # it was a deletion.
                if adjustedOffset in otherBases:
                    if otherBases[adjustedOffset] is None:
                        assert offset in actualDeletions
                else:
                    otherBases[adjustedOffset] = (
                        reference.sequence[offset]
                        if noCoverage == "reference"
                        else noCoverage
                    )

            bar.update(barCount)

    return correspondences, consensusBases, otherBases, originalOffsets


def pileupCorrespondences(bam, bamId, referenceLength, includeSoftClipped, progress):
    """
    Get bases for offsets that correspond to the reference using pysam's
    pileup funciton.

    @param bam: An open BAM file.
    @param bamId: A C{str} reference name indicating which reference to
        find reads against in the BAM file.
    @param referenceLength: The C{int} length of the reference.
    @param includeSoftClipped: Include information from read bases that were
        marked as soft-clipped by the algorithm that made the BAM file.
    @param progress: If C{True}, display a progress bar on standard error.
    @return: A C{dict} keyed by C{int} reference offset, with C{Bases}
        instances as values.
    """
    correspondences = defaultdict(lambda: Bases())

    with maybeProgressBar(progress, referenceLength, "Pileup   : ") as bar:
        for barCount, site in enumerate(bam.pileup(contig=bamId), start=1):
            referenceOffset = site.reference_pos
            for read in site.pileups:
                if not includeSoftClipped:
                    cigarOperations = list(
                        cigarTuplesToOperations(read.alignment.cigartuples)
                    )
                queryOffset = read.query_position
                if queryOffset is None or (
                    not includeSoftClipped
                    and cigarOperations[queryOffset] == CSOFT_CLIP
                ):
                    continue
                    # Make sure an empty Bases instance is placed in our
                    # results so there's a sign that the reference offset
                    # was returned by bam.pileup.
                    correspondences[referenceOffset]
                else:
                    alignment = read.alignment
                    correspondences[referenceOffset].append(
                        alignment.query_sequence[queryOffset],
                        alignment.query_qualities[queryOffset],
                    )

            bar.update(barCount)

    return correspondences


def compareCorrespondences(fp, fetch, pileup, threshold, minCoverage):
    """
    Compare bases at corresponding offsets as obtained from the pysam fetch
    and pileup functions.

    @param fp: An open file pointer to write the summary to.
    @param fetch: A defaultdict(lambda: Bases()) keyed by C{int} offset.
    @param pileup: A defaultdict(lambda: Bases()) keyed by C{int} offset.
    @param threshold: A C{float} threshold, as in C{consensusFromBAM}.
    @param minCoverage: An C{int} minimum number of reads that must cover a
        site for a consensus base to be called.
    """
    fetchOffsets = set(offset for offset in fetch if offset >= 0)
    pileupOffsets = set(pileup)

    unmatched = pileupOffsets - fetchOffsets
    if unmatched:
        print(
            f"There are {len(unmatched)} sites obtained from pileup not "
            f"present in fetch:",
            file=fp,
        )

        for offset in sorted(unmatched):
            print(f"  {offset + 1}: {pileup[offset]}", file=fp)

    unmatched = fetchOffsets - pileupOffsets
    if unmatched:
        print(
            f"There are {len(unmatched)} sites obtained from fetch not "
            f"present in pileup:",
            file=fp,
        )

        for offset in sorted(unmatched):
            print(f"  {offset + 1}: {fetch[offset]}", file=fp)

    common = fetchOffsets & pileupOffsets
    if common:
        print(
            f"There are {len(common)} sites obtained from both fetch and pileup.",
            file=fp,
        )
        differenceCount = 0
        for offset in sorted(common):
            fetchBase = fetch[offset].consensus(
                threshold, minCoverage, "LOW-COVERAGE", "NO-COVERAGE"
            )
            pileupBase = pileup[offset].consensus(
                threshold, minCoverage, "LOW-COVERAGE", "NO-COVERAGE"
            )
            if fetchBase != pileupBase:
                differenceCount += 1
                print(f"  {offset + 1}: {fetchBase} != {pileupBase}", file=fp)
                if fetch[offset]:
                    print(f"    Fetch : {fetch[offset]}", file=fp)
                if pileup[offset]:
                    print(f"    Pileup: {pileup[offset]}", file=fp)

        if differenceCount:
            print(
                f"There were differences called at "
                f"{pct(differenceCount, len(common))} sites and "
                f"{pct(len(common) - differenceCount, len(common))} were "
                f"called identically.",
                file=fp,
            )
        else:
            print(f"All {len(common)} sites were called identically.", file=fp)
